#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **mapping merger** (i.e., :class:`set`-style union and intersection
operations on dictionary-like types or instances) functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.enums import make_enum
from betse.util.type.types import type_check, MappingType, IterableTypes
from copy import deepcopy

# ....................{ ENUMERATIONS                      }....................
MergeCollisionPolicy = make_enum(
    class_name='MergeCollisionPolicy',
    member_names=(
        'RAISE_EXCEPTION',
        'PREFER_FIRST',
        'PREFER_LAST',
    ),
)
'''
Enumeration of all supported types of **key collision merger policies** (i.e.,
strategies for merging keys shared between two or more mappings).

Key collisions occur when two or more mappings to be merged contain key-value
pairs containing the same keys but differing values -- or, equivalently, when a
key of a key-value pair of one mapping is also a key of a key-value pair of
another mapping such that the two values of these two pairs differ.

Attributes
----------
RAISE_EXCEPTION : enum
    This policy raises a fatal exception on the first key collision. This
    constitutes the strictest and hence safest such policy.
PREFER_FIRST : enum
    This policy accepts *only* the first key-value pair with this key in these
    mappings and thus ignores all subsequent such pairs in subsequent mappings
    on every key collision. This policy gives higher precedence to keys in
    mappings passed earlier -- the converse of the :attr:`PREFER_LAST` policy.
PREFER_LAST : enum
    This policy accepts *only* the last key-value pair with this key in these
    mappings and thus ignores all prior such pairs in prior mappings on every
    key collision. This policy gives higher precedence to keys in mappings
    passed later -- the converse of the :attr:`PREFER_FIRST` policy.
'''

# ....................{ MERGERS                           }....................
@type_check
def merge_maps(
    # Mandatory parameters.
    mappings: IterableTypes,

    # Optional parameters.
    collision_policy: MergeCollisionPolicy = (
        MergeCollisionPolicy.RAISE_EXCEPTION),
    is_values_copied: bool = False,
) -> MappingType:
    '''
    Dictionary of all key-value pairs merged together from all passed
    dictionaries (in the passed order) with the passed key collision policy.

    Caveats
    ----------
    All classes of all passed dictionaries *must* define ``__init__()`` methods
    accepting a dictionary comprehension.

    Whether the order in which dictionaries are passed is significant
    conditionally depends on the passed ``collision_policy`` parameter.
    Specifically, if this parameter is:

    * :attr:`MergeCollisionPolicy.RAISE_EXCEPTION`, **order is insignificant.**
      In this case, this function simply raises an exception if any two of the
      passed dictionaries **key-collide** (i.e., define key-value pairs of the
      same key but differing values).
    * :attr:`MergeCollisionPolicy.PREFER_FIRST`, **order is significant.** When
      one or more mappings to be merged contain the same key, this function
      accepts only the first key-value pair with this key in these mappings and
      thus ignores all subsequent such pairs in subsequent mappings.
    * :attr:`MergeCollisionPolicy.PREFER_LAST`, **order is significant.** When
      one or more mappings to be merged contain the same key, this function
      accepts only the last key-value pair with this key in these mappings and
      thus ignores all prior such pairs in prior mappings.

    Parameters
    ----------
    mappings : IterableTypes
        Iterable of all dictionaries to be merged.
    collision_policy : optional[MergeCollisionPolicy]
        **Key collision policy** (i.e., strategy for merging keys shared by
        one or more mappings) to apply. Defaults to
        :attr:`MergeCollisionPolicy.RAISE_EXCEPTION`, raising an exception on
        the first key collision.
    is_values_copied : optional[bool]
        Either:

        * ``True``, if all values of the merged dictionary returned by this
          function are deeply (i.e., recursively) copied from their source
          dictionaries.
        * ``False``, if all values of the merged dictionary returned by this
          function are shallowly (i.e., non-recursively) copied from their
          source dictionaries.

        Defaults to ``False``.

    Returns
    ----------
    MappingType
        Dictionary merged from and of the same type as the passed dictionaries.

    Raises
    ----------
    BetseMappingException
        If the passed ``collision_policy`` is
        :attr:`MergeCollisionPolicy.RAISE_EXCEPTION` *and* any key of any
        key-value pair of any passed dictionary is also a key of any
        key-value pair of other such dictionary whose values differ.
    BetseSequenceException
        If less than two mappings are passed.

    See Also
    ----------
    http://treyhunner.com/2016/02/how-to-merge-dictionaries-in-python
        Blog post strongly inspiring this implementation. Thanks, Trey!
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import itertest, sequences
    from betse.util.type.iterable.mapping import maptest

    # Sequence of mappings converted from this iterable of mappings.
    mappings = sequences.to_sequence(mappings)

    # If less than two mappings were passed, raise an exception.
    sequences.die_if_length_less_than(sequence=mappings, length=2)
    # Else, at least two mappings were passed.

    # If any passed mapping is *NOT* a mapping, raise an exception.
    itertest.die_unless_items_instance_of(iterable=mappings, cls=MappingType)

    # Type of dictionary to be returned.
    dict_type = type(mappings[0])

    # Function accepting a single value and returning either a shallow or deep
    # copy of that value depending on which the caller requested.
    copy_value = deepcopy if is_values_copied else lambda value: value

    # If raising an exception on the first key collision, do so *BEFORE*
    # performing this merger.
    #
    # While there exist a countably infinite number of approaches to merging
    # non-colliding dictionaries in Python, this is the optimally efficient.
    # The is_maps_collide() function underlying the die_if_maps_collide()
    # function called below reduces to two symmetric set differences of
    # arbitrarily many iterables. Likewise, the merger performed below
    # transparently supports this key collision policy assuming no collisions.
    if collision_policy is MergeCollisionPolicy.RAISE_EXCEPTION:
        maptest.die_if_maps_collide(*mappings)
    # If giving higher precedence to dictionaries passed earlier, reverse the
    # order of the passed dictionaries. Why? Because the algorithm implemented
    # below implements the "PREFER_LAST" rather than "PREFER_FIRST" policy by
    # default. Since the former is merely the reverse of the latter, reversing
    # the order of these dictionaries repeatedly replaces the subsequent value
    # of each colliding key from the subsequent dictionary with the prior value
    # of that key in the prior dictionary, efficiently implementing the
    # "PREFER_FIRST" policy.
    elif collision_policy is MergeCollisionPolicy.PREFER_FIRST:
        mappings = reversed(mappings)

    # Dictionary merged from the passed dictionaries via a doubly-nested
    # dictionary comprehension (in the passed order). This repeatedly replaces
    # the prior value of each colliding key from the prior dictionary with the
    # subsequent value of that key in the subsequent dictionary, efficiently
    # implementing the "PREFER_LAST" policy.
    #
    # While there exist a countably infinite number of approaches to merging
    # dictionaries in Python, this approach is known to be the most efficient
    # for general-purpose merging of arbitrarily many dictionaries under Python
    # >= 3.4. See also Trey Hunter's exhaustive commentary (complete with
    # timings) at the above URL.
    dict_merged = {
        key: copy_value(value)
        for mapping in mappings
        for key, value in mapping.items()
    }

    # Return a dictionary of this type converted from this dictionary. If the
    # desired type is a "dict", this dictionary is returned as is; else, this
    # dictionary is converted into an instance of the desired type.
    return dict_merged if dict_type is dict else dict_type(dict_merged)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level logging configuration classes.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific
# modules should *ALWAYS* be able to safely import from this module at any
# scoping level, circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import logging, os, sys
from beartype import beartype
from betse.util.io.log.logenum import LogLevel
from betse.util.type.types import type_check
from logging import (
    Handler,
    RootLogger,
    StreamHandler,
)

# ....................{ CONFIG                             }....................
#FIXME: Update docstring to reflect the new default configuration.
class LogConf(object):
    '''
    Application-specific logging configuration.

    This configuration defines sane default filters, formatters, and handlers
    for the root logger, which callers may customize (e.g., according to
    user-defined settings) by setting various properties of this configuration.

    Caveats
    ----------
    **The :meth:`__init__` method may raise exceptions.** Hence, this class
    should be instantiated at application startup by an explicit call to the
    :func:`betse.util.io.log.conf.logconf.init` function *after* establishing
    default exception handling.

    **The :meth:`deinit` method must be called at application shutdown.** Doing
    so closes the logfile handle opened by the :meth:`__init__` method. Failure
    to do so will reliably raise non-fatal warnings (hidden by default, but
    visible while running tests) resembling:

        C:\\projects\\betse\\betse\\util\\io\\log\\conf\\logconf.py:44:
        ResourceWarning: unclosed file <_io.TextIOWrapper
        name='C:\\Users\\appveyor\\AppData\\Roaming\\betse\\betse.log' mode='a'
        encoding='utf-8'>
            _log_conf = LogConf()

    Default Settings
    ----------
    All loggers will implicitly propagate messages to the root logger
    configured by this class, whose output will be:

    * Formatted in a timestamped manner detailing the point of origin (e.g.,
      ``[2016-04-03 22:02:47] betse ERROR (util.py:50): File not found.``).
    * Labelled as the current logger's name, defaulting to `root`. Since this
      is *not* a terribly descriptive name, callers are encouraged to replace
      this by an application-specific name.
    * Printed to standard error if the logging level for this output is either
      ``WARNING``, ``ERROR``, or ``CRITICAL``.
    * Printed to standard output if the logging level for this output is
      ``INFO``. Together with the prior item, this suggests that output with a
      logging level of ``DEBUG`` will *not* be printed by default.
    * Appended to the user-specific file defined by the
      :meth:`app_meta.log_default_filename` property, whose:

      * Level defaults to :data:`logger.ALL`. Hence, *all* messages will be
        logged by default, including low-level debug messages. (This is
        helpful for debugging client-side errors.)
      * Contents will be automatically rotated on exceeding a sensible filesize
        (e.g., 16Kb).

    If the default log levels are undesirable, consider subsequently calling
    this logger's :meth:`Logger.setLevel` method. Since a desired log level is
    typically unavailable until after parsing CLI arguments and/or
    configuration file settings *and* since a logger is required before this
    level becomes available, this class provides a sane interim default.

    Attributes
    ----------
    _filename : str
        Absolute or relative path of the file logged to by the file handler,
        defaulting to :meth:`app_meta.log_default_filename`.
    _logger_root : Logger
        Root logger.
    _logger_root_handler_file : Handler
        Root logger handler appending to the current logfile.
    _logger_root_handler_stderr : Handler
        Root logger handler printing to standard error.
    _logger_root_handler_stdout : Handler
        Root logger handler printing to standard output.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self):
        '''
        Initialize this logging configuration.

        Specifically, this method associates the root logger with a rotating
        logfile whose handle remains open until the :meth:`deinit` method is
        subsequently called at application shutdown.
        '''

        # Avoid circular import dependencies.
        from betse.util.test import tsttest

        # Initialize the superclass.
        super().__init__()

        # Initialize all non-property attributes to sane defaults. To avoid
        # chicken-and-egg issues, properties should *NOT* be set here.
        self._deinit_vars()

        # Initialize the root logger.
        self._init_logger_root()

        # Initialize root logger handlers *AFTER* the root logger, as the
        # former explicitly add themselves to the latter.
        self._init_logger_root_handler_std()
        self._init_logger_root_handler_file()

        # Redirect all warnings through the logging framewark *AFTER*
        # successfully performing the above initialization.
        logging.captureWarnings(True)

        # If the active Python interpreter is running an automated test suite,
        # manually increase logging verbosity as soon as feasible: i.e., here.
        # While deferring this configuration to elsewhere (namely, the
        # "betse_test.fixture.initter" submodule) is also feasible, doing so
        # would horrifyingly squelch all early-time debug messages.
        if tsttest.is_testing():
            self.is_verbose = True


    def _init_logger_root(self) -> None:
        '''
        Initialize the root logger.

        For safety, this function removes all previously initialized handlers
        from this logger.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone

        # Root logger.
        self._logger_root = logging.getLogger()

        # For uniqueness, change the name of the root logger to that of our
        # top-level package "betse" from its ambiguous default "root".
        self._logger_root.name = appmetaone.get_app_meta().package_name

        # Instruct this logger to entertain all log requests, ensuring these
        # requests will be delegated to the handlers defined below. By default,
        # this logger ignores all log requests with level less than "WARNING",
        # preventing handlers from receiving these requests.
        self._logger_root.setLevel(LogLevel.ALL)

        # Safely remove all existing handlers from the root logger *BEFORE*
        # adding new handlers to this logger.
        #
        # The root logger should have no handlers under conventional use cases.
        # The root logger only has handlers for functional tests *NOT*
        # parallelized by "xdist" and hence running in the same Python process.
        self._deinit_logger_root_handlers()


    def _init_logger_root_handler_std(self) -> None:
        '''
        Initialize root logger handlers redirecting log messages to the
        standard stdout and stderr file handles.
        '''

        # Avoid circular import dependencies.
        from betse.util.io.log.logfilter import (
            LogFilterThirdPartyDebug, LogFilterMoreThanInfo)
        from betse.util.io.log.conf.logconfformat import LogFormatterWrap
        from betse.util.os.command import cmds

        # Initialize the stdout handler to:
        #
        # * Log only informational messages by default.
        # * Unconditionally ignore all warning and error messages, which the
        #   stderr handler already logs.
        #
        # Sadly, the "StreamHandler" constructor does *NOT* accept the
        # customary "level" attribute accepted by its superclass constructor.
        self._logger_root_handler_stdout = StreamHandler(sys.stdout)
        self._logger_root_handler_stdout.setLevel(LogLevel.INFO)
        self._logger_root_handler_stdout.addFilter(LogFilterMoreThanInfo())

        # Initialize the stderr handler to:
        #
        # * Log only warning and error messages by default.
        # * Unconditionally ignore all informational and debug messages, which
        #   the stdout handler already logs.
        self._logger_root_handler_stderr = StreamHandler(sys.stderr)
        self._logger_root_handler_stderr.setLevel(LogLevel.WARNING)

        # Avoid printing third-party debug messages to the terminal.
        self._logger_root_handler_stdout.addFilter(LogFilterThirdPartyDebug())
        self._logger_root_handler_stderr.addFilter(LogFilterThirdPartyDebug())

        #FIXME: Consider colourizing this format string.

        # Format standard output and error in the conventional way. For a list
        # of all available log record attributes, see:
        #
        #     https://docs.python.org/3/library/logging.html#logrecord-attributes
        #
        # Note that "{{" and "}}" substrings in format() strings escape literal
        # "{" and "}" characters, respectively.
        stream_format = '[{}] {{message}}'.format(cmds.get_current_basename())

        # Formatter for this format.
        stream_formatter = LogFormatterWrap(fmt=stream_format, style='{')

        # Assign these formatters to these handlers.
        self._logger_root_handler_stdout.setFormatter(stream_formatter)
        self._logger_root_handler_stderr.setFormatter(stream_formatter)

        # Register these handlers with the root logger.
        self._logger_root.addHandler(self._logger_root_handler_stdout)
        self._logger_root.addHandler(self._logger_root_handler_stderr)


    def _init_logger_root_handler_file(self) -> None:
        '''
        Initialize the root logger handler appending log messages to the
        currently open logfile file handle.

        This method is designed to be called multiple times, permitting the
        filename associated with this handler to be modified at runtime.
        '''

        # Avoid circular import dependencies.
        from betse.util.io.log.logfilter import LogFilterThirdPartyDebug
        from betse.util.io.log.conf.logconfformat import LogFormatterWrap
        from betse.util.io.log.conf.logconfhandle import (
            LogHandlerFileRotateSafe)
        from betse.util.path import pathnames
        from betse.util.os.command import cmds
        from betse.util.type.numeric import ints

        # Absolute or relative path of the directory containing this file.
        file_dirname = pathnames.get_dirname(self._filename)

        # Minimum level of messages to be log to disk, defaulting to "INFO".
        file_level = LogLevel.INFO

        # If this handler has already been created...
        if self._logger_root_handler_file is not None:
            # Preserve the previously set minimum level of messages to log.
            file_level = self._logger_root_handler_file.level

            # If the root logger has also already been created, deinitialize
            # this handler from this root logger.
            if self._logger_root is not None:
                self._deinit_logger_root_handler(
                    self._logger_root_handler_file)

        # If the dirname of the directory containing this file is non-empty,
        # create this directory if needed. Note this dirname is empty when this
        # filename is a pure basename (e.g., when the "--log-file=my.log"
        # option is passed).
        #
        # For safety, this directory is created with standard low-level Python
        # functionality rather than our custom higher-level
        # dirs.make_parent_unless_dir() function. The latter logs this
        # creation. Since the root logger is *NOT* fully configured yet,
        # calling that function here would induce subtle errors or exceptions.
        if file_dirname:
            os.makedirs(file_dirname, exist_ok=True)

        # Root logger file handler, preconfigured as documented above.
        self._logger_root_handler_file = LogHandlerFileRotateSafe(
            filename=self._filename,

            # Append rather than overwrite this file.
            mode='a',

            # Defer opening this file in a just-in-time manner (i.e., until the
            # first call to this handler's emit() method is called to write the
            # first log via this handler). Why? Because (in no particular
            # order):
            #
            # * If the end user requests that *NO* messages be logged to disk
            #   (e.g., by passing the "--log-level=none" option), this file
            #   should *NEVER* be opened and hence created. The simplest means
            #   of doing so is simply to indefinitely defer opening this file.
            # * Doing so slightly reduces the likelihood (but *NOT* eliminate
            #   the possibility) of race conditions between multiple BETSE
            #   processes attempting to concurrently rotate the same logfile.
            delay=True,

            # Encode this file's contents as UTF-8.
            encoding='utf-8',

            # Maximum filesize in bytes at which to rotate this file,
            # equivalent to 1MB.
            maxBytes=ints.MiB,

            # Maximum number of rotated logfiles to maintain.
            backupCount=8,
        )

        # Initialize this handler's level to the previously established level.
        self._logger_root_handler_file.setLevel(file_level)

        # Prevent third-party debug messages from being logged to disk.
        self._logger_root_handler_file.addFilter(LogFilterThirdPartyDebug())

        # Linux-style logfile format.
        #
        # Note that the "processName" attribute appears to *ALWAYS* expand to
        # "MainProcess", which is not terribly descriptive. Hence, the name of
        # the current process is manually embedded in this format.
        file_format = (
            '[{{asctime}}] '
            '{} {{levelname}} '
            '({{module}}.py:{{funcName}}():{{lineno}}) '
            '<PID {{process}}>:\n'
            '    {{message}}'.format(cmds.get_current_basename()))

        # Format this file according to this format.
        file_formatter = LogFormatterWrap(fmt=file_format, style='{')
        self._logger_root_handler_file.setFormatter(file_formatter)

        # Register this handler with the root logger.
        self._logger_root.addHandler(self._logger_root_handler_file)

    # ..................{ DEINITIALIZERS                     }..................
    def deinit(self) -> None:
        '''
        Deinitialize this logging configuration.

        See Also
        ----------
        :func:`_deinit_logger_root_handlers`
        :func:`_deinit_vars`
            Further details.
        '''

        # Deinitialize all root logger handlers.
        self._deinit_logger_root_handlers()

        # Deinitialize all instance variables.
        self._deinit_vars()


    def _deinit_logger_root_handlers(self) -> None:
        '''
        Deinitialize all root logger handlers.

        Specifically, this method iterates over all handlers previously added
        to the root logger and, for each such handler (in order):

        #. Closes all open file handles associated with that handler, including
           the logfile handle opened by the :meth:`__init__` method.
        #. Removes that handler from the root logger.
        '''

        # For each handler previously added to the root logger, deinitialize
        # that handler.
        #
        # For safety, a shallow copy of the list of handlers to be removed
        # rather than the actual list being modified here is iterated over.
        for root_handler in tuple(self._logger_root.handlers):
            self._deinit_logger_root_handler(root_handler)


    @beartype
    def _deinit_logger_root_handler(self, root_handler: Handler) -> None:
        '''
        Deinitialize the passed root logger handler.

        Specifically, this method:

        #. Closes all open file handles associated with this handler, including
           the logfile handle opened by the :meth:`__init__` method.
        #. Removes this handler from the root logger.
        '''

        # Close all open file handles associated with this handler.
        root_handler.close()

        # Remove this handler from the root logger.
        self._logger_root.removeHandler(root_handler)


    def _deinit_vars(self) -> None:
        '''
        Deinitialize all instance variables underlying this logging
        configuration to sane defaults.

        To circumvent chicken-and-egg issues, this method intentionally avoids
        deinitializing (i.e., setting) settable properties.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone

        # Revert all non-property attributes to sane defaults.
        self._filename = appmetaone.get_app_meta().log_default_filename
        self._logger_root = None
        self._logger_root_handler_file = None
        self._logger_root_handler_stderr = None
        self._logger_root_handler_stdout = None

    # ..................{ PROPERTIES ~ logger                }..................
    # Read-only properties prohibiting write access to external callers.

    @property
    def logger_root(self) -> RootLogger:
        '''
        **Root logger** (i.e., transitive parent of all other loggers).
        '''

        return self._logger_root

    # ..................{ PROPERTIES ~ handler               }..................
    @property
    def handler_file(self) -> Handler:
        '''
        Root logger handler appending to the current logfile if file logging is
        enabled *or* ``None`` otherwise.
        '''

        return self._logger_root_handler_file


    @property
    def handler_stderr(self) -> Handler:
        '''
        Root logger handler printing to standard error.
        '''

        return self._logger_root_handler_stderr


    @property
    def handler_stdout(self) -> Handler:
        '''
        Root logger handler printing to standard output.
        '''

        return self._logger_root_handler_stdout

    # ..................{ PROPERTIES ~ level                 }..................
    @property
    def file_level(self) -> LogLevel:
        '''
        Minimum level of messages to log to the file handler.
        '''

        return self._logger_root_handler_file.level


    @file_level.setter
    @type_check
    def file_level(self, file_level: LogLevel) -> None:
        '''
        Set the minimum level of messages to log to the file handler.
        '''

        self._logger_root_handler_file.setLevel(file_level)

    # ..................{ PROPERTIES ~ level : verbose       }..................
    @property
    def is_verbose(self) -> bool:
        '''
        ``True`` only if *all* messages are to be unconditionally logged to the
        stdout handler (and hence printed to stdout).

        Equivalently, this method returns ``True`` only if the logging level
        for the stdout handler is :attr:`LogLevel.ALL`.

        Note that this logging level is publicly retrievable by accessing the
        :attr:`handler_stdout.level` property.
        '''

        return self._logger_root_handler_stdout.level == LogLevel.ALL


    @is_verbose.setter
    @type_check
    def is_verbose(self, is_verbose: bool) -> None:
        '''
        Set the verbosity of the stdout handler.

        This method sets this handler's logging level to:

        * If the passed boolean is ``True``, :attr:`LogLevel.ALL` .
        * If the passed boolean is ``False``, :attr:`LogLevel.INFO`.
        '''

        # Avoid circular import dependencies.
        from betse.util.io.log import logs

        # Convert the passed boolean to a logging level for the stdout handler.
        self._logger_root_handler_stdout.setLevel(
            LogLevel.ALL if is_verbose else LogLevel.INFO)

        # If increasing stdout verbosity, log this fact *AFTER* doing so.
        #
        # Note that reversing this order of statements would silently squelch
        # this message -- which would quite defeat the purpose.
        if is_verbose:
            logs.log_debug('Standard output verbosity enabled.')

    # ..................{ PROPERTIES ~ path                  }..................
    @property
    def filename(self) -> str:
        '''
        Absolute or relative path of the file logged to by the file handler.
        '''

        return self._filename


    @filename.setter
    @type_check
    def filename(self, filename: str) -> None:
        '''
        Set the absolute or relative path of the file logged to by the file
        handler.

        Due to flaws in the upstream :mod:`logging` API, this method
        necessarily destroys and recreates the current file handler.
        '''

        # If the passed filename is the same as the current filename, avoid
        # unnecessarily destroying and recreating the file handler. This is
        # technically a negligible optimization, but every little bit helps.
        if self._filename == filename:
            return

        # Classify this filename *BEFORE* recreating the file handler, which
        # accesses this variable.
        self._filename = filename

        # Destroy and recreate the file handler.
        self._init_logger_root_handler_file()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed gene regulatory network (GRN) subconfigurations.
'''

# ....................{ IMPORTS                           }....................
from betse.lib.yaml.yamlalias import yaml_alias
from betse.lib.yaml.abc.yamlfileabc import YamlFileABC
# from betse.science.config.grn.confgrnyadda import SimConfGrnYadda
from betse.util.type.types import type_check  #, MappingType, SequenceTypes

# ....................{ SUBCLASSES                        }....................
class SimConfGrnFile(YamlFileABC):
    '''
    YAML-backed in-memory and on-disk gene regulatory network (GRN)
    subconfiguration, encapsulating a low-level container of *all* GRN-related
    configuration settings both loaded from and saved back to a YAML-formatted
    configuration file.

    Attributes (General)
    ----------
    '''

    # ..................{ ALIASES                           }..................
    # is_overlay_current = yaml_alias(
    #     "['results options']['overlay currents']", bool)

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Encapsulate low-level dictionaries with high-level wrappers.
        # self.anim_while_sim = SimConfExportAnimCellsEmbedded()

        # Encapsulate low-level lists of dictionaries with high-level wrappers.
        # self.anims_after_sim = SimConfExportVisualCells.make_list()

    # ..................{ LOADERS                           }..................
    def load(self, *args, **kwargs) -> None:

        # Load our superclass with all passed arguments.
        super().load(*args, **kwargs)

        # Load all subconfigurations of this configuration.
        # self.anim_while_sim.load(conf=self._conf[
        #     'results options']['while solving']['animations'])


    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Unload all subconfigurations of this configuration.
        # self.anim_while_sim.unload()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixtures creating temporary directories isolated for safety to specific tests.
'''

# ....................{ IMPORTS                            }....................
from pytest import fixture

# ....................{ FIXTURES                           }....................
# Test-scope fixture creating and returning a new object for each unique test.
@fixture
def betse_temp_dir(
    request: '_pytest.python.FixtureRequest',
    tmpdir_factory: '_pytest.tmpdir.tmpdir_factory',
) -> 'py._path.local.LocalPath':
    '''
    Per-test fixture creating a temporary directory isolated to the test
    requesting this fixture and returning an object encapsulating this
    directory.

    This directory is guaranteed to be specific to this test. The basename of
    this directory is the name of this test excluding the prefixing substring
    ``test_``. When requested by the ``test_cli_sim_default`` test, for
    example, this fixture creates a temporary directory
    ``{tmpdir}/cli_sim_default`` for the absolute path ``{tmpdir}`` of this
    test session's root temporary directory (e.g.,
    ``/tmp/pytest-0/cli_sim_default``).

    This directory is safely accessible *only* for the duration of this test.
    Subsequently run tests and fixtures *cannot* safely reuse this directory,
    although doing so is technically feasible in unreliable edge-cases.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Builtin fixture describing the parent fixture or test of this fixture.
    tmpdir_factory : _pytest.tmpdir.tmpdir_factory
        Builtin session-scoped fixture whose ``mktemp()`` method returns a
        :class:`py.path.local` instance encapsulating a new temporary
        directory.

    Returns
    ----------
    py._path.local.LocalPath
        Object encapsulating this temporary directory.
    '''

    # Defer heavyweight imports.
    from betse.util.test.pytest import pytrequest
    from betse.util.type.text.string import strs

    # Name of the current test.
    test_name = pytrequest.get_tested_name(request)
    # print('    request.node: {}'.format(request.node))
    # print('    test_name: {}'.format(test_name))

    # Basename of the temporary directory containing this configuration file,
    # set to the name of the current test excluding the prefixing "test_".
    temp_dir_basename = strs.remove_prefix(
        text=test_name,
        prefix='test_',
        exception_message=(
            f'Test name "{test_name}" not prefixed by "test_".'),
    )

    # Create this temporary directory and wrap this directory's absolute path
    # with a high-level "py.path.local" object. See also:
    #     http://pytest.org/latest/tmpdir.html#the-tmpdir-factory-fixture
    temp_dirpath = tmpdir_factory.mktemp(temp_dir_basename)

    # Return this object.
    return temp_dirpath



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **testing-specific continuous integration** (i.e., remote host
exercising all functional and unit tests within this application's test suite
across one or more Python versions and one or more platforms) facilities.

See Also
----------
https://media.readthedocs.org/pdf/python-semantic-release/latest/python-semantic-release.pdf
    Section 3.2.1, *Environment checks,* from the official documentation for
    the completely unrelated albeit significantly more popular
    ``python-semantic-version`` project. This section reliably documents the
    set of all environment variables uniquely identifying continuous
    integration (CI) hosts, thus serving as the principal inspiration for this
    whole submodule. Thanks for all the CI, ``python-semantic-version``.
'''

# ....................{ IMPORTS                           }....................
# from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ TESTERS                           }....................
def is_ci_circle() -> bool:
    '''
    ``True`` only if the active Python interpreter is currently running tests
    under the remote CircleCI continuous integration (CI) host.
    '''

    return _is_ci_host(env_var_name='CIRCLECI')


def is_ci_frigg() -> bool:
    '''
    ``True`` only if the active Python interpreter is currently running tests
    under the remote Frigg continuous integration (CI) host.
    '''

    return _is_ci_host(env_var_name='FRIGG')


def is_ci_gitlab() -> bool:
    '''
    ``True`` only if the active Python interpreter is currently running tests
    under a remote GitLab-CI continuous integration (CI) host.
    '''

    return _is_ci_host(env_var_name='GITLAB_CI')


def is_ci_semaphore() -> bool:
    '''
    ``True`` only if the active Python interpreter is currently running tests
    under the remote Semphare continuous integration (CI) host.
    '''

    return _is_ci_host(env_var_name='SEMAPHORE')


def is_ci_travis() -> bool:
    '''
    ``True`` only if the active Python interpreter is currently running tests
    under the remote Travis-CI continuous integration (CI) host.
    '''

    return _is_ci_host(env_var_name='TRAVIS')

# ....................{ TESTERS ~ private                 }....................
@type_check
def _is_ci_host(env_var_name: str) -> bool:
    '''
    ``True`` only if the active Python interpreter is currently running tests
    (e.g., with the :mod:`pytest` test harness) *and* an environment variable
    with the passed name uniquely associated with one remote continuous
    integration (CI) host exists in the shell environment encapsulating this
    interpreter, implying this host to be running this interpreter and hence
    these tests.

    Parameters
    ----------
    env_var_name : str
        Name of the environment variable uniquely identifying this interpreter
        to be running under one remote continuous integration (CI) host.

    Returns
    ----------
    bool
        ``True`` only if this interpreter is running tests under the CI host
        uniquely identified by this variable.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.shell import shellenv

    # Return true only if this interpreter is running tests *AND* an
    # environment variable with the passed name exists, implying these tests to
    # be run under the CI host uniquely identified with that variable.
    #
    # Ideally, this implementation would also call the
    # betse.util.test.tests.is_calling() function to ensure that tests are
    # running as expected. However, that function is *NOT* safely callable from
    # module scope. Calling that function here would thus prevent this function
    # and functions calling this function from also being safely callable from
    # module scope -- which, in turn, would prevent test decorators (e.g.,
    # @betse.util.test.pytest.mark.pytskip.skip_if_ci_gitlab) from being safely
    # used... *AT ALL.* Since that would rather defeat the purpose of defining
    # any of this functionality in the first place, we have little choice but
    # to defer to a simple environment variable existence test.
    return shellenv.is_var(env_var_name)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

r'''
Low-level \*ML (e.g., HTML, SGML, XML) facilities.
'''

# ....................{ IMPORTS                            }....................
import html
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check, RegexCompiledType

# ....................{ GETTERS                            }....................
@func_cached
def get_tag_regex() -> RegexCompiledType:
    r'''
    Compiled regular expression matching a syntactically but *not* necessarily
    semantically valid \*ML (e.g., HTML, SGML, XML) tag.

    For efficiency in downstream clients (e.g., BETSEE) frequently calling the
    :func:`is_ml` function and hence requiring this expression be compiled, this
    expression is intentionally pre-compiled rather than uncompiled and thus
    returned as a cached getter.

    See Also
    ----------
    https://kevin.deldycke.com/2008/07/python-ultimate-regular-expression-to-catch-html-tags
        Blog article strongly inspiring this regular expression.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Create, return, and cache this expression.
    return regexes.compile_regex(
        # Tag prefix.
        r'<'
        # Tag closure prefix (optional).
        r'\/?'
        # Tag name.
        r'\w+'
        # Zero or more tag attributes.
        r'(?:'
            # Attribute name.
            r'\s+\w+'
            # Attribute value (optional).
            r'(?:'
                # Equals sign.
                r'\s*=\s*'
                # Either:
                r'(?:'
                    # Double-quoted string.
                    r'".*?"|'
                    # Single-quoted string.
                    r"'.*?'|"
                    # Unquoted word.
                    r"[^'"
                    r'">\s]+'
                r')'
            r')?'
        r')*'
        # Whitespace (optional).
        r'\s*'
        # Tag closure suffix (optional).
        r'\/?'
        # Tag suffix.
        r'>'
    )

# ....................{ TESTERS                            }....................
@type_check
def is_ml(text: str) -> bool:
    r'''
    ``True`` only if the passed string contains one or more \*ML tags and hence
    superficially appears to contain \*ML-specific syntax.

    Parameters
    ----------
    text : str
        String to be tested for the presence of \*ML-specific syntax.

    Returns
    ----------
    bool
        ``True`` only if this string contains one or more \*ML tags.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Test this string against this expression.
    return regexes.is_match(text=text, regex=get_tag_regex())

# ....................{ ESCAPERS                           }....................
@type_check
def escape_ml(text: str) -> str:
    r'''
    Passed string with all \*ML-specific syntax escaped and hence guaranteed to
    contain *no* \*ML-specific syntax.

    Specifically, this function converts all:

    * ``<`` characters into ``&lt;`` substrings.
    * ``>`` characters into ``&gt;`` substrings.
    * ``&`` characters into ``&amp;`` substrings.
    * ``'`` characters into ``&#x27;`` substrings.
    * ``"`` characters into ``&quot;`` substrings.

    Caveats
    ----------
    This function renders arbitrary strings syntactically but *not* necessarily
    semantically suitable for use as valid \*ML (e.g., HTML, SGML, XML).
    Notably, returned strings may still contain unescaped JavaScript and hence
    cross-site scripting (XSS) vulnerabilities. These strings should *never* be
    directly displayed as HTML to end users. For example, the following call to
    this function preserves a JavaScript fragment unescaped:

        >>> from betse.util.type.text import mls
        >>> mls.escape_ml('<a href="javascript:alert()">')
        '&lt;a href=&quot;javascript:alert()&quot;&gt;'

    Parameters
    ----------
    text : str
        String to escape all \*ML-specific syntax of.

    Returns
    ----------
    str
        String with all \*ML-specific syntax escaped.
    '''

    # Praise be the stdlib.
    return html.escape(text)

# ....................{ TAGIFIERS                          }....................
@type_check
def tagify_newlines(text: str) -> str:
    '''
    Convert each newline in the passed string into a ``<br/>`` tag.

    Parameters
    ----------
    text : str
        String to convert all newlines of.

    Returns
    ----------
    str
        String with all newlines converted to corresponding tags.
    '''

    # Sometimes, it really is just that easy.
    return text.replace('\n', '<br/>')



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Voltage-gated calcium channel classes.
'''

# .................... IMPORTS                            ....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.science import sim_toolbox as stb
from betse.science.channels.channelsabc import ChannelsABC
from betse.science.math import toolbox as tb
from betse.util.io.log import logs

# .................... BASE                               ....................
class VgCaABC(ChannelsABC, metaclass=ABCMeta):
    '''
    Abstract base class of all Voltage-gated calcium channel classes.

    Attributes
    ----------
    _mInf :
        Equillibrium value function for m-gates of channel
    _mTau :
        Time-constant function for m-gates of channel
    _hInf :
        Equillibrium value function for h-gates of channel
    _hTau :
        Time-constant function for h-gates of channel
    '''

    def init(self, vm, cells, p, targets = None):
        '''
        Initialize targeted voltage-gated calcium channels at the initial
        time step of the simulation based on the initial cell Vmems.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.
        '''

        if targets is None:

            self.targets = None

            if type(vm) == float or type(vm) == np.float64:
                self.data_length = 1

            else:
                self.data_length = len(vm)


        else:
            self.targets = targets
            self.data_length = len(self.targets)
            self.mdl = len(cells.mem_i)

        self.modulator = 1.0

        self.v_corr = 0.0  # in experiments, the measurement junction voltage is about 10 mV

        if self.targets is None:

            V = vm * 1000 + self.v_corr

        else:
            V = vm[self.targets] * 1000 + self.v_corr

        self._init_state(V)

        self.ions = ['Ca']
        self.rel_perm = [1.0]


    def run(self, vm, p):
        '''
        Handle all targeted voltage-gated sodium channels by working with the passed
        user-specified parameters on the tissue simulation and cellular
        world for a time step.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.

        '''
        if self.targets is None:

            V = vm*1000 + self.v_corr

        else:

            V = vm[self.targets] * 1000 + self.v_corr

        self._calculate_state(V)

        self._implement_state(V, p)

    def _implement_state(self, V, p):

        # Update the 'm' and 'h' gating functions:
        self.update_mh(p, time_unit = self.time_unit)

        # calculate the open-probability of the channel:
        P = (self.m ** self._mpower) * (self.h ** self._hpower)

        if self.targets is None:

            self.P = P

        else:
            self.P = np.zeros(self.mdl)
            self.P[self.targets] = P


    @abstractmethod
    def _init_state(self, V):
        '''
        Do something.
        '''
        pass


    @abstractmethod
    def _calculate_state(self, V):
        '''
        Do something.
        '''
        pass

# ....................{ SUBCLASS                           }....................

class Cav3p3(VgCaABC):
    '''

    T-type calcium channel model from Traboulsie et al. Based on a Cav3.3 channel.


    Reference: Traboulsie A. et al. Subunit-specific modulation of T-type calcium
    channels by zinc. J. Physiol. (Lond.), 2007 Jan 1 , 578 (159-71).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Cav3p3')

        self.time_unit = 1.0e3

        self.vrev = 30     # reversal voltage used in model [mV]
        Texpt = 0.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp((V - -45.454426) / -5.073015))
        self.h = 1 / (1 + np.exp((V - (-74.031965)) / 8.416382))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1 / (1 + np.exp((V - -45.454426) / -5.073015))
        self._mTau = 3.394938 + (54.187616 / (1 + np.exp((V - -40.040397) / 4.110392)))
        self._hInf = 1 / (1 + np.exp((V - (-74.031965)) / 8.416382))
        self._hTau = 109.701136 + (0.003816 * np.exp(-V / 4.781719))

class Cav2p1(VgCaABC):
    '''

    P/Q-type calcium channel (Cav2.1) model from Miyasho et al.

    Found in the nervous system, especially Purkunjie neurons,  play an important role in mediating neurotransmitter
    release in the nervous system, postsynaptic integration, neuroplasticity, neural excitability,
    and gene transcription.


    Reference: Miyasho T. et al. Low-threshold potassium channels and a low-threshold calcium channel
    regulate Ca2+ spike firing in the dendrites of cerebellar Purkinje neurons: a modeling study.
    Brain Res., 2001 Feb 9 , 891 (106-15).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Cav2p1')

        self.time_unit = 1.0e3

        self.vrev = 135.0     # reversal voltage used in model [mV]
        Texpt = 0.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0  # FIXME implement this!

        mAlpha = 8.5 / (1 + np.exp((V - 8) / (-12.5)))
        mBeta = 35 / (1 + np.exp((V + 74) / (14.5)))

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = mAlpha / (mAlpha + mBeta)
        self.h = 1.0

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        mAlpha = 8.5 / (1 + np.exp((V - 8) / (-12.5)))
        mBeta = 35 / (1 + np.exp((V + 74) / (14.5)))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)

        self._hInf = 1.0
        self._hTau = 1.0

class Cav1p3(VgCaABC):
    '''

    Low voltage activating L-type calcium channel model from Avery et al.


    Reference: Avery RB. et al. Multiple channel types contribute to the low-voltage-activated calcium current in
    hippocampal CA3 pyramidal neurons. J. Neurosci., 1996 Sep 15 , 16 (5567-82).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Cav1.3')

        self.time_unit = 1.0e3

        self.vrev = 113.0     # reversal voltage used in model [mV]
        Texpt = 21.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)

        # V = V - 10

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self.h = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self._mTau = (5.0000 + 20.0000 / (1 + np.exp((V - -25.000) / 5)))
        self._hInf = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))
        self._hTau = (20.0000 + 50.0000 / (1 + np.exp((V - -40.000) / 7)))

class Cav1p2(VgCaABC):
    '''

    L-type calcium channel model of Carlin et. al.

    Reference: Carlin KP. et al. Characterization of calcium
    currents in functionally mature mouse spinal motoneurons. Eur. J. Neurosci., 2000 May , 12 (1624-34).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Cav1.2')

        self.time_unit = 1.0e3

        self.vrev = 131.0     # reversal voltage used in model [mV]
        Texpt = 21.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)

        V = V - 10

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self.h = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        V = V - 10
        self._mInf = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self._mTau = (5.0000 + 20.0000 / (1 + np.exp((V - -25.000) / 5)))
        self._hInf = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))
        self._hTau = (20.0000 + 50.0000 / (1 + np.exp((V - -40.000) / 7)))

class Cav2p3(VgCaABC):
    '''

    R-type calcium channel model from Miyasho et al. based on Cav2p3.


    Reference: Miyasho T. et al. Low-threshold potassium channels and a low-threshold calcium channel
    regulate Ca2+ spike firing in the dendrites of cerebellar Purkinje neurons:
    a modeling study. Brain Res., 2001 Feb 9 , 891 (106-15).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the R-type vgCa channel type: Cav2p3')

        self.time_unit = 1.0e3

        self.vrev = 30     # reversal voltage used in model [mV]
        Texpt = 0.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0  # FIXME implement this!

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        mAlpha = 2.6 / (1 + np.exp((V + 7) / -8))
        mBeta = 0.18 / (1 + np.exp((V + 26) / 4))
        hAlpha = 0.0025 / (1 + np.exp((V + 32) / 8))
        hBeta = 0.19 / (1 + np.exp((V + 42) / -10))

        self.m = mAlpha / (mAlpha + mBeta)
        self.h = hAlpha / (hAlpha + hBeta)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        mAlpha = 2.6 / (1 + np.exp((V + 7) / -8))
        mBeta = 0.18 / (1 + np.exp((V + 26) / 4))
        hAlpha = 0.0025 / (1 + np.exp((V + 32) / 8))
        hBeta = 0.19 / (1 + np.exp((V + 42) / -10))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)
        self._hInf =hAlpha / (hAlpha + hBeta)
        self._hTau = 1 / (hAlpha + hBeta)

class Cav2p2(VgCaABC):
    '''

    N-type calcium channel model from Huang SJ et al.


    Reference: Huang SJ. et al. Activation and inactivation properties of voltage-gated calcium currents
    in developing cat retinal ganglion cells. Neuroscience, 1998 Jul , 85 (239-47).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the N-type vgCa channel type: Cav2p2')

        self.time_unit = 1.0e3

        self.vrev = 135     # reversal voltage used in model [mV]
        Texpt = 36.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0

        # indsV = (V == 20.0).nonzero()
        # V[indsV] += 1.0e-6

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        mAlpha = (0.1 * (V - 20) / (1 - np.exp(-(V - 20) / 10)))
        mBeta = 0.4 * np.exp(-(V + 25) / 18)
        hAlpha = 0.01 * np.exp(-(V + 50) / 10)
        hBeta =  0.1 / (1 + np.exp(-(V + 17) / 17))


        self.m = mAlpha / (mAlpha + mBeta)
        self.h = hAlpha / (hAlpha + hBeta)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        # indsV = (V == 20.0).nonzero()
        # V[indsV] += 1.0e-6

        mAlpha = (0.1 * (V - 20) / (1 - np.exp(-(V - 20) / 10)))
        mBeta = 0.4 * np.exp(-(V + 25) / 18)
        hAlpha = 0.01 * np.exp(-(V + 50) / 10)
        hBeta =  0.1 / (1 + np.exp(-(V + 17) / 17))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)
        self._hInf =hAlpha / (hAlpha + hBeta)
        self._hTau = 1 / (hAlpha + hBeta)

class Cav3p1(VgCaABC):
    '''

    T-type calcium channel model from Traboulsie et al, based on Cav3p1.


    Reference: Traboulsie A. et al. Subunit-specific modulation of T-type calcium channels by zinc.
    J. Physiol. (Lond.), 2007 Jan 1 , 578 (159-71).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the T-type vgCa channel type: Cav3p1')

        self.time_unit = 1.0e3

        self.vrev = 30     # reversal voltage used in model [mV]
        Texpt = 0.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0  # FIXME implement this!

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp((V - (-42.921064)) / -5.163208))
        self.h = 1 / (1 + np.exp((V - (-72.907420)) / 4.575763))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1 / (1 + np.exp((V - (-42.921064)) / -5.163208))

        # inds_lt10 = (V < -10e-3).nonzero()
        inds_gt10 = (V >= -10e-3).nonzero()

        self._mTau = -0.855809 + (1.493527 * np.exp(-V / 27.414182))
        self._mTau[inds_gt10] = 1.0
        self._hInf = 1 / (1 + np.exp((V - (-72.907420)) / 4.575763))
        self._hTau = 9.987873 + (0.002883 * np.exp(-V / 5.598574))

class Ca_L2(VgCaABC):
    '''

    L-type calcium channel model Avery et al. L-type channels are higher-voltage activating and very
    persistent. They are commonly found in muscle or glands, where they induce activities such as hormone
    release or muscle contraction in response to neural stimulation.

    This channel has been modified to activate at more depolarized Vmem.


    Reference: Avery RB. et al. Multiple channel types contribute to the low-voltage-activated calcium current in
    hippocampal CA3 pyramidal neurons. J. Neurosci., 1996 Sep 15 , 16 (5567-82).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Ca_L2')

        self.time_unit = 1.0e3

        self.vrev = 131.0     # reversal voltage used in model [mV]
        Texpt = 36.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self.h = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self._mTau = 10.0
        self._hInf = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))
        self._hTau = 59.0

class Ca_L3(VgCaABC):
    '''

    L-type calcium channel model Avery et al. L-type channels are higher-voltage activating and very
    persistent. They are commonly found in muscle or glands, where they induce activities such as hormone
    release or muscle contraction in response to neural stimulation.

    This channel has been modified to activate at more depolarized Vmem and to have a broader activation peak.


    Reference: Avery RB. et al. Multiple channel types contribute to the low-voltage-activated calcium current in
    hippocampal CA3 pyramidal neurons. J. Neurosci., 1996 Sep 15 , 16 (5567-82).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Ca_L3')

        self.time_unit = 1.0e3

        self.vrev = 131.0     # reversal voltage used in model [mV]
        Texpt = 36.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0

        V = V - 15

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self.h = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        # self._mInf = 1.0000 / (1 + np.exp(((V - 10) + 30.000) / -6))
        # self._mTau = 5.0000 + 20.0000 / (1 + np.exp(((V - 10) + 25.000) / 5))
        # self._hInf = 1.0000 / (1 + np.exp(((V - 10) + 80.000) / 6.4))
        # self._hTau = 20.0000 + 50.0000 / (1 + np.exp(((V - 10) + 40.000) / 7))

        V = V - 15

        self._mInf = 1.0000 / (1 + np.exp((V - -30.000) / -6))
        self._mTau = 10.0
        self._hInf = 1.0000 / (1 + np.exp((V - -80.000) / 6.4))
        self._hTau = 59.0

class Cav_G(VgCaABC):
    '''

    General calcium channel model from Reuveni et al.

    Reference: Reuveni I. et al. Stepwise repolarization from Ca2+ plateaus in neocortical pyramidal cells: evidence
    for nonhomogeneous distribution of HVA Ca2+ channels in dendrites. J. Neurosci., 1993 Nov , 13 (4609-21).




    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgCa channel type: Cav_G')

        self.time_unit = 1.0e3

        self.vrev = 131.0     # reversal voltage used in model [mV]
        Texpt = 36.0    # temperature of the model in degrees C
        # self.qt = 2.3**((simT-Texpt)/10)
        self.qt = 1.0


        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        mAlpha = (0.055 * (-27 - V)) / (np.exp((-27 - V) / 3.8) - 1)
        mBeta = (0.94 * np.exp((-75 - V) / 17))
        self.m = mAlpha / (mAlpha + mBeta)

        hAlpha = (0.000457 * np.exp((-13 - V) / 50))
        hBeta = (0.0065 / (np.exp((-V - 15) / 28) + 1))
        self.h = hAlpha / (hAlpha + hBeta)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """


        mAlpha = (0.055 * (-27 - V)) / (np.exp((-27 - V) / 3.8) - 1)
        mBeta = (0.94 * np.exp((-75 - V) / 17))
        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)
        hAlpha = (0.000457 * np.exp((-13 - V) / 50))
        hBeta = (0.0065 / (np.exp((-V - 15) / 28) + 1))
        self._hInf = hAlpha / (hAlpha + hBeta)
        self._hTau = 1 / (hAlpha + hBeta)

class CaLeak(VgCaABC):

    '''
    Simple calcium leak channel -- always open -- for substance modulation.

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using a substance-modulated Ca++ channel')

        self.time_unit = 1.0

        self.vrev = 30     # reversal voltage used in model [mV]
        Texpt = 23    # temperature of the model in degrees C
        self.qt = 1.0

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = np.ones(self.data_length)
        self.h = np.ones(self.data_length)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 0
        self._hpower = 0


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1
        self._mTau = 1
        self._hInf = 1
        self._hTau = 1







#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level :mod:`matplotlib`-specific figure functionality.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To allow matplotlib defaults (e.g., backend, logging) to be replaced
# with application-specific preferences, the unsafe "matplotlib.pyplot"
# submodule must *NOT* be imported here at the top-level.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.exceptions import BetseMatplotlibException
from betse.util.io.log import logs
from betse.util.type.types import type_check, MatplotlibFigureType

# ....................{ EXCEPTIONS                         }....................
def die_unless_figure() -> None:
    '''
    Raise an exception if *no* figures are currently open with the
    :mod:`matplotlib.pyplot` GCF API.

    Raises
    ------
    BetseMatplotlibException
        If no figures are currently open with this API.

    See Also
    --------
    :func:`is_figure`
        Further details.
    '''

    if not is_figure():
        raise BetseMatplotlibException('No matplotlib figures currently open.')

# ....................{ TESTERS                            }....................
def is_figure() -> bool:
    '''
    ``True`` only if one or more figures are currently open with the
    :mod:`matplotlib.pyplot` GCF API.
    '''

    # Defer heavyweight imports.
    from matplotlib import pyplot

    # Note that, despite the name, the pyplot.get_fignums() function returns a
    # list of the numbers of currently open figures rather than the number of
    # currently open figures. Welcome to Matplotlib Hell, where all is well.
    return bool(pyplot.get_fignums())

# ....................{ GETTERS                            }....................
def get_figure_current() -> MatplotlibFigureType:
    '''
    Figure most recently opened with the :mod:`matplotlib.pyplot` GCF API.

    Specifically, this function returns the same figure instance returned by
    the most recent call to the :func:`matplotlib.pyplot.figure` function.
    '''

    # Defer heavyweight imports.
    from matplotlib import pyplot

    # If no figure is currently open, raise an exception.
    die_unless_figure()

    # Three-letter acronyms do *NOT* constitute a human-readable API.
    return pyplot.gcf()

# ....................{ CLOSERS                            }....................
@type_check
def close_figure(figure: MatplotlibFigureType) -> None:
    '''
    **Close** (i.e., clear, delete, remove, garbage collect) the passed figure,
    guaranteeing that all resources consumed by this figure will be
    subsequently reclaimed after an indeterminate period of time.

    Specifically, this function (in order):

    #. Nullifies the contents of all non-axes artists of this figure.
    #. Nullifies the contents of all axes of this figure.
    #. Closes all interactive windows (if any) associated with this figure.
    #. Deassociates this figure from the :mod:`matplotlib.pyplot` GCF API.

    Caveats
    -----------
    **Figure closure is surprisingly non-trivial.** Failure to call this
    function on open figures usually results in some or all of the contents of
    those figures continuing to reside in memory. Ergo, this high-level
    function should *always* be called in lieu of low-level
    :mod:`matplotlib.pyplot` functions (e.g., :mod:`matplotlib.pyplot.close`)
    or equally low-level :class:`matplotlib.figure.Figure` methods (e.g.,
    :meth:`matplotlib.figure.Figure.cla`), which are all known to behave
    non-deterministically and hence unsafely.

    Parameters
    -----------
    figure : MatplotlibFigureType
        Figure to be closed.

    See Also
    -----------
    https://stackoverflow.com/a/17106460/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Defer heavyweight imports.
    from matplotlib import pyplot

    # Log this closure.
    logs.log_debug('Closing matplotlib figure "%r"...', figure.number)

    # Nullify the contents of all non-axes artists of this figure.
    figure.clf()

    # For each axis of this figure, nullify the contents of this axis *AFTER*
    # that of non-axes. Why? We have utterly no idea. The matplotlib API is
    # dark voodoo that, frankly, not even matplotlib developers grok. It is
    # unsafe to ask too many questions.
    for figure_axis in figure.axes:
        figure_axis.cla()

    # Deassociate this figure from the "matplotlib.pyplot" GCF API, implicitly
    # closing all interactive windows (if any) associated with this figure. In
    # theory, doing so should also implicitly nullify the above content; in
    # practice, however, the matplotlib API is sufficiently unreliable across
    # version bumps that explicitly nullifying that content is all but
    # non-optional. Thanks for all of the bugs, matplotlib!
    pyplot.close(figure)


def close_figure_current() -> None:
    '''
    **Close** (i.e., clear, delete, remove, garbage collect) the figure most
    recently opened with the :mod:`matplotlib.pyplot` GCF API, guaranteeing
    that all resources consumed by this figure will be subsequently reclaimed
    after an indeterminate period of time.

    Specifically, this function closes the same figure instance returned by
    the most recent call to the :func:`matplotlib.pyplot.figure` function.

    See Also
    -----------
    :func:`close_figure`
        Further details.
    '''

    # Stay classy, matplotlib.
    close_figure(get_figure_current())


def close_figures_all() -> None:
    '''
    **Close** (i.e., clear, delete, remove, garbage collect) all figures
    currently opened with the :mod:`matplotlib.pyplot` GCF API, guaranteeing
    that all resources consumed by these figures will be subsequently reclaimed
    after an indeterminate period of time.

    See Also
    -----------
    https://stackoverflow.com/a/13174720/2809027
        StackOverflow answer strongly inspiring this implementation.
    :func:`close_figure`
        Further details.
    '''

    # Defer heavyweight imports.
    from matplotlib import pyplot

    # Log this closure.
    logs.log_debug('Closing all matplotlib figures...')

    # Mortifying API, thy name is the "pyplot" interface.
    pyplot.close('all')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Simulation phase callback** (i.e., caller-defined object whose methods are
periodically called while simulating one or more phases) class hierarchy.
'''

# ....................{ IMPORTS                           }....................
from betse.util.io.log import logs
from betse.util.test import tsttest
from betse.util.type.call.callbacks import CallbacksBC
from betse.util.type.types import NoneType, StrOrNoneTypes

# ....................{ SUPERCLASSES                      }....................
# This subclass is currently an empty placeholder but will be subsequently
# extended with subclass-specific behaviour.
class SimCallbacksBC(CallbacksBC):
    '''
    Abstract base class of all **simulation phase callbacks** (i.e.,
    caller-defined object whose methods are periodically called while
    simulating one or more simulation phases) subclasses.
    '''

    pass

# ....................{ TYPES                             }....................
SimCallbacksBCOrNoneTypes = (SimCallbacksBC, NoneType)
'''
Tuple of both the simulation phase callbacks type *and* that of the ``None``
singleton.
'''

# ....................{ SUBCLASSES                        }....................
class SimCallbacksNoop(SimCallbacksBC):
    '''
    **Noop simulation phase callbacks** (i.e., simulation phase callbacks whose
    methods all silently reduce to noops for efficiency).

    This callbacks subclass is typically instantiated as a convenience fallback
    in the event that no other callbacks subclass is passed by an external
    caller to a simulation subcommand. For efficiency, all callbacks
    reimplemented by this subclass intentionally:

    * Do *not* type-check any parameters passed by the source callable.
    * Do *not* call their superclass implementations.

    See Also
    ----------
    :func:`make_default`
        Factory function internally instantiating and returning a new instance
        of this subclass, which should typically be called in lieu of manually
        instantiating instances of this subclass. This function handles edge
        cases intentionally excluded by this subclass, including the need to
        instantiate an instance of a different subclass when running tests.
    '''

    # ..................{ CALLBACKS                         }..................
    def progress_ranged(
        self, progress_max: int, progress_min: int = 0) -> None:
        pass

    def progress_stated(self, status: str) -> None:
        pass

    def progressed(self, progress: int) -> None:
        pass

    def progressed_last(self, status: StrOrNoneTypes = None) -> None:
        pass

    def progressed_next(self, status: StrOrNoneTypes = None) -> None:
        pass

# ....................{ MAKERS                            }....................
def make_default() -> SimCallbacksBC:
    '''
    Create and return a new simulation phase callbacks object suitable for use
    as an efficient fallback in the event that a caller fails to supply a
    caller-specific callbacks object.

    Specifically, this factory function:

    * If tests are currently being run, this function creates and returns an
      instance of the :class:`SimCallbacksBC` subclass. While less efficient
      than the comparable :class:`SimCallbacksNoop` subclass, doing so ensures
      that contractual guarantees maintained by the :class:`CallbacksBC`
      superclass are properly exercised.
    * Else, this function creates and returns an instance of the
      :class:`SimCallbacksNoop` subclass. While more efficient than the
      comparable :class:`SimCallbacksBC` subclass, doing so avoids contractual
      guarantees maintained by the :class:`CallbacksBC` superclass.
    '''

    # If tests are currently being run...
    if tsttest.is_testing():
        # Log this behaviour.
        logs.log_debug('Enabling test-specific callbacks...')

        # Return a new simulation phase callbacks object preserving (and hence
        # testing) superclass API guarantees.
        return SimCallbacksBC()
    # Else, tests are *NOT* currently being run. In this case...
    else:
        # Log this behaviour.
        logs.log_debug('Enabling optimized noop callbacks...')

        # Return a new simulation phase callbacks object maximizing efficiency.
        return SimCallbacksNoop()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level classes aggregating all parameters pertaining to simulation events.
'''

# ....................{ IMPORTS                           }....................
from betse.science.tissue.event.tiseveabc import SimEventSpikeABC
# from betse.util.io.log import logs
from betse.util.type.types import type_check  #, SequenceTypes

# ....................{ SUBCLASSES                        }....................
#FIXME: Refactor the TissueHandler.removeCells() method into a new
#fire() method of this subclass.
class SimEventCut(SimEventSpikeABC):
    '''
    **Cutting event** (i.e., event removing one or more cells of the current
    cell cluster at some time step during the simulation phase).
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize this cutting event for the passed simulation configuration.

        Attributes
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # Initialize our superclass.
        super().__init__(p=p, time_step=p.event_cut_time)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation subconfiguration classes for tissue and cut profiles.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta
from betse.lib.yaml.yamlalias import (
    yaml_alias,
    yaml_alias_float_nonnegative,
    yaml_alias_float_percent,
    yaml_enum_alias,
)
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.lib.yaml.abc.yamllistabc import YamlList, YamlListItemABC
from betse.lib.yaml.abc.yamlmixin import YamlNamedMixin
from betse.science.enum.enumconf import CellsPickerType
# from betse.util.io.log import logs
from betse.util.type.types import type_check, SequenceTypes

# ....................{ SUPERCLASSES ~ tissue             }....................
class SimConfTissueABC(YamlNamedMixin, metaclass=ABCMeta):
    '''
    Mixin of all **YAML-backed tissue profile subconfiguration** (i.e.,
    configuration of a single tissue profile parsed from the current
    YAML-formatted simulation configuration file) subclasses.

    Attributes (Membrane Diffusion)
    ----------
    Dm_Na : float
        Sodium (Na+) membrane diffusion constant in m2/s.
    Dm_K : float
        Potassium (K+) membrane diffusion constant in m2/s.
    Dm_Cl : float
        Chloride (Cl-) membrane diffusion constant in m2/s.
    Dm_Ca : float
        Calcium (Ca2+) membrane diffusion constant in m2/s.
    Dm_M : float
        Charge balance anion (M-) membrane diffusion constant in m2/s.
    Dm_P : float
        Protein (P-) membrane diffusion constant in m2/s.
    '''

    # ..................{ ALIASES ~ diffusion               }..................
    Dm_Na = yaml_alias_float_nonnegative("['diffusion constants']['Dm_Na']")
    Dm_K  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_K']")
    Dm_Cl = yaml_alias_float_nonnegative("['diffusion constants']['Dm_Cl']")
    Dm_Ca = yaml_alias_float_nonnegative("['diffusion constants']['Dm_Ca']")
    Dm_M  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_M']")
    Dm_P  = yaml_alias_float_nonnegative("['diffusion constants']['Dm_P']")

# ....................{ SUBCLASSES ~ tissue               }....................
class SimConfTissueDefault(SimConfTissueABC, YamlABC):
    '''
    YAML-backed **default tissue profile** (i.e., profile applied to all cells
    *not* already targeted by another tissue profile) subconfiguration,
    encapsulating the configuration of a single tissue profile unconditionally
    applicable to all cells parsed from a dictionary configuring at least this
    profile in the current YAML-formatted simulation configuration file.

    Attributes (Cell Picker)
    ----------
    picker_image_filename : str
        Absolute or relative filename of the image mask whose pure-black pixels
        define the shape of the cell cluster to be populated with cells. See
        the :attr:`SimConfTissueListItem.picker_image_filename` variable for
        details.
    '''

    # ..................{ ALIASES ~ picker                  }..................
    picker_image_filename = yaml_alias("['image']", str)


class SimConfTissueListItem(SimConfTissueABC, YamlListItemABC):
    '''
    YAML-backed tissue profile list item subconfiguration, encapsulating the
    configuration of a single tissue profile parsed from a list of these
    profiles in the current YAML-formatted simulation configuration file.

    Attributes
    ----------
    is_gj_insular : bool
        ``True`` only if gap junctions originating at cells in this tissue are
        **insular** (i.e., prevented from connecting to cells in other
        tissues), implying these gap junctions to be strictly intra-tissue.

    Attributes (Cell Picker)
    ----------
    picker_type : CellsPickerType
        Type of **tissue profile picker** (i.e., object assigning a subset of
        all cells matching some criteria to this tissue profile).
    picker_cells_color : str
        **Hexadecimal-formatted color** (i.e., string of six hexadecimal digits
        specifying this color's red, green, and blue components) of all circles
        within the vector image (defined by the ``cells from svg`` setting in
        the current simulation configuration) to be assigned to this tissue.
        Ignored unless :attr:`picker_type` is :attr:`CellsPickerType.COLOR`.
    picker_cells_index : SequenceTypes
        Sequence of the indices of all cells to be assigned to this tissue.
        Ignored unless :attr:`picker_type` is :attr:`CellsPickerType.INDICES`.
    picker_cells_percent : float
        **Percentage** (i.e., floating point number in the range ``[0.0,
        100.0]``) of the total cell population to be randomly assigned to this
        tissue. Ignored unless :attr:`picker_type` is
        :attr:`CellsPickerType.PERCENT`.
    picker_image_filename : str
        Absolute or relative filename of the image mask whose pure-black pixels
        (i.e., pixels whose red, green, and blue color components are all 0)
        define the region of the cell cluster whose cells are all to be
        assigned to this tissue. This image *must*:

        * Be square (i.e., have equal width and height).
        * Contain no alpha transparency layer.

        Ignored unless :attr:`picker_type` is :attr:`CellsPickerType.IMAGE`.
    '''

    # ..................{ ALIASES                           }..................
    is_gj_insular = yaml_alias("['insular']", bool)

    # ..................{ ALIASES ~ picker                  }..................
    picker_type = yaml_enum_alias("['cell targets']['type']", CellsPickerType)

    #FIXME: Create a new yaml_alias_color() data descriptor validating this
    #string to be a valid hexadecimal-formatted color.
    #FIXME: In the YAML, if color specifications don't have quotations around
    #them (e.g. '808080') the yaml sometimes thinks all numerical hex codes are
    #strings (e.g. 008080), and sometimes integers (e.g. 800080), in which case
    #the whole thing crashed with a type error! I'm not quite sure how to force
    #it to read in a string... For now, I've put quotes around the hex code
    #numeral...
    picker_cells_color = yaml_alias("['cell targets']['color']", str)

    picker_cells_index = yaml_alias(
        "['cell targets']['indices']", SequenceTypes)
    picker_cells_percent = yaml_alias_float_percent(
        "['cell targets']['percent']")
    picker_image_filename = yaml_alias(
        "['cell targets']['image']", str)

    # ..................{ SUPERCLASS                        }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the first tissue profile in our default configuration file.
        return cls._make_loaded(conf={
            'name': yaml_list.get_item_name_uniquified('Tissue ({})'),
            'insular': True,
            'diffusion constants': {
                'Dm_Na': 1.0e-18,
                'Dm_K':  15.0e-18,
                'Dm_Cl': 2.0e-18,
                'Dm_Ca': 1.0e-18,
                'Dm_M':  1.0e-18,
                'Dm_P':  0.0,
            },
            'cell targets': {
                'type': 'all',
                'color': 'ff0000',
                'image': 'geo/circle/circle_base.png',
                'indices': [3, 14, 15, 9, 265],
                'percent': 50,
            },
        })

# ....................{ SUBCLASSES ~ cut                  }....................
class SimConfCutListItem(YamlNamedMixin, YamlListItemABC):
    '''
    YAML-backed cut profile list item subconfiguration, encapsulating the
    configuration of a single cut profile parsed from a list of these profiles
    in the current YAML-formatted simulation configuration file.

    Attributes
    ----------
    picker_image_filename : str
        Absolute or relative filename of the image mask whose pure-black pixels
        define the region of the cell cluster whose cells are all to be removed
        by this cut profile. See the
        :attr:`SimConfTissueListItem.picker_image_filename` variable for
        details.
    '''

    # ..................{ ALIASES                           }..................
    picker_image_filename = yaml_alias("['image']", str)

    # ..................{ SUPERCLASS                        }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the first tissue profile in our default configuration file.
        return cls._make_loaded(conf={
            'name': yaml_list.get_item_name_uniquified('Cut ({})'),
            'image': 'geo/circle/wedge.png',
        })



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **set** (i.e., :class:`set`-like types or instances) functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.types import type_check, IterableTypes, SetType

# ....................{ MAKERS                            }....................
@type_check
def make_union(*iterables: IterableTypes) -> SetType:
    '''
    **Union** (i.e., set of all objects contained in all passed iterables) of
    all passed iterables.

    This function generalizes the :meth:`set.union` method, which implicitly
    requires the first iterable to be a set, to arbitrary iterables, none of
    which (including the first iterable) are required to be a set.

    Parameters
    ----------
    iterables : Tuple[IterableTypes]
        Tuple of all iterables to be united.

    Raises
    ----------
    BetseSequenceException
        If less than two iterables are passed.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import generators, sequences

    # If passed no iterables, raise an exception.
    sequences.die_if_empty(sequence=iterables)
    # Else, at least one iterable is passed.

    # If passed only one generator, implicitly expand this generator into the
    # sequence of all items yielded by this generator.
    if len(iterables) == 1 and generators.is_generator(iterables[0]):
        iterables = tuple(iterables[0])

    # First iterable.
    iterable_first = iterables[0]

    # First iterable coerced into a set if nor already a set *OR* preserved as
    # is otherwise.
    #
    # Note that testing whether this iterable is a "SetType" does *NOT*
    # suffice, as that abstract base class sadly fails to require that
    # subclasses adhere to the public API of the "set" builtin (e.g., by
    # defining a union() method).
    iterable_first_set = (
        iterable_first if isinstance(iterable_first, set) else
        set(iterable_first))

    # If passed only one iterable, simply return this set.
    if len(iterables) == 1:
        return iterable_first_set
    # Else, two or more iterables are passed.

    # Generator comprehension yielding each subsequent iterable.
    iterables_rest = (iterable for iterable in iterables[1:])

    # Union of all passed iterables.
    iterables_united = iterable_first_set.union(*iterables_rest)

    # Return this union.
    return iterables_united



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level class facilities.
'''

# ....................{ IMPORTS                           }....................
import inspect
from betse.exceptions import BetseTypeException
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ClassType,
    GeneratorType,
    MappingOrNoneTypes,
    SequenceOrNoneTypes,
    StrOrNoneTypes,
)

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_subclass(subclass: ClassType, superclass: ClassType) -> None:
    '''
    Raise an exception unless the first passed class inherits and is thus a
    subclass of the second passed class.

    For simplicity, note that any class is considered to be a subclass of
    itself within the context of this function. (See the example below.)

    Parameters
    ----------
    subclass : ClassType
        Subclass to be validated.
    superclass : ClassType
        Superclass to be validated.

    Raises
    ----------
    BetseTypeException
        If the first such class is *not* a subclass of the second such class.

    Examples
    ----------
        >>> from betse.util.type.cls import classes
        >>> class antiihnii(object): pass
        >>> class yee_naaldlooshii(antiihnii): pass
        >>> classes.die_unless_subclass(
        ...     subclass=antiihnii, superclass=antiihnii)
        >>> classes.die_unless_subclass(
        ...     subclass=yee_naaldlooshii, superclass=yee_naaldlooshii)
        >>> classes.die_unless_subclass(
        ...     subclass=yee_naaldlooshii, superclass=antiihnii)
        >>> classes.die_unless_subclass(
        ...     subclass=antiihnii, superclass=yee_naaldlooshii)
        Traceback (most recent call last):
          File "/home/leycec/tmp/yilo.py", line 12, in <module>
            subclass=antiihnii, superclass=yee_naaldlooshii)
          File "<string>", line 24, in __die_unless_subclass_type_checked__
          File "/home/leycec/py/betse/betse/util/type/cls/classes.py", line 63, in die_unless_subclass
            subclass, superclass))
        betse.exceptions.BetseTypeException: Class <class '__main__.antiihnii'> not a subclass of class <class '__main__.yee_naaldlooshii'>
    '''

    if not issubclass(subclass, superclass):
        raise BetseTypeException(
            'Class {!r} not a subclass of class {!r}'.format(
                subclass, superclass))

# ....................{ TESTERS                           }....................
@type_check
def is_abstract(cls: ClassType) -> bool:
    '''
    ``True`` only if the passed class is an **abstract base class** (i.e., a
    class whose metaclass is :class:`abc.ABCMeta` such that at least one
    abstract method decorated by the :func:`abc.abstractmethod` decorator
    remains unimplemented).

    Parameters
    ----------
    cls : ClassType
        Class to be tested.

    Returns
    ----------
    bool
        ``True`` only if this class is an abstract base class (ABC).
    '''

    # One-liners for ponderous victory.
    return inspect.isabstract(cls)

# ....................{ GETTERS                           }....................
@type_check
def get_module_name_qualified(cls: ClassType) -> str:
    '''
    Fully-qualified name of the module defining the passed class if this class
    is defined by a module *or* raise an exception otherwise.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this module name for.

    Returns
    ----------
    str
        Fully-qualified name of this module.

    Raises
    ----------
    BetseTypeException
        If this class is *not* defined by a module. Specifically, if either:

        * This class defines no special ``__module__`` attribute.
        * This class defines a special ``__module__`` attribute whose value is
          either:

          * ``None``.
          * A special identifier. Notably, builtin classes defined by no
            modules (e.g., :class:`str`) define a special ``__module__``
            attribute whose value is ``__builtin__``.

    See Also
    ----------
    https://stackoverflow.com/a/13653312/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyident

    # If this class does *NOT* define the special "__module__" attribute,
    # raise an exception. In theory, all classes including builtin classes
    # should define this. In practice, some do not. Thanks alot, everybody!
    if not hasattr(cls, '__module__'):
        raise BetseTypeException(
            'Class "{class_name}" module undefined '
            '(i.e., "{class_name}.__module__" attribute not found).'.format(
                class_name=get_name_unqualified(cls)))
    # Else, this class defines this attribute.

    # Fully-qualified name of the module defining this class if any *OR* "None"
    # otherwise.
    module_name = cls.__module__

    # If this name is either undefined *OR* prefixed and suffixed by both "__"
    # (e.g., as is the case with "str.__module__ == '__builtin__'"), no module
    # defines this class. In this case, raise an exception.
    if module_name is None or pyident.is_special(module_name):
        raise BetseTypeException(
            'Class "{class_name}" module undefined '
            '(i.e., "{class_name}.__module__" attribute is '
            '"{module_name}").'.format(
                class_name=get_name_unqualified(cls),
                module_name=str(module_name)))

    # Else, return this name.
    return module_name


@type_check
def get_module_name_qualified_or_none(cls: ClassType) -> StrOrNoneTypes:
    '''
    Fully-qualified name of the module defining the passed class if this class
    is defined by a module *or* ``None`` otherwise.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this module name for.

    Returns
    ----------
    StrOrNoneTypes
        Fully-qualified name of this module if any *or* ``None`` otherwise.
        Specifically, ``None`` is returned if either:

        * This class defines no special ``__module__`` attribute.
        * This class defines a special ``__module__`` attribute whose value is
          either:

          * ``None``.
          * A special identifier. Notably, builtin classes defined by no
            modules (e.g., :class:`str`) define a special ``__module__``
            attribute whose value is ``__builtin__``.

    See Also
    ----------
    https://stackoverflow.com/a/13653312/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyident

    # If this class does *NOT* define the special "__module__" attribute,
    # return "None". In theory, all classes including builtin classes should
    # define this. In practice, some do not. Thanks for nuthin', somebody!
    if not hasattr(cls, '__module__'):
        return None
    # Else, this class defines this attribute.

    # Fully-qualified name of the module defining this class if any *OR* "None"
    # otherwise.
    module_name = cls.__module__

    # Return "None" if this name is either undefined *OR* prefixed and suffixed
    # by both "__" (e.g., as is the case with "str.__module__ ==
    # '__builtin__'") *OR* this name as is otherwise.
    return (
        None
        if module_name is None or pyident.is_special(module_name) else
        module_name
    )

# ....................{ GETTERS ~ name                    }....................
@type_check
def get_name_qualified(cls: ClassType) -> str:
    '''
    Fully-qualified name of the passed class.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this fully-qualified name for.

    Returns
    ----------
    str
        Fully-qualified name of this class.
    '''

    # Unqualified name of this class.
    class_name = get_name_unqualified(cls)

    # Fully-qualified name of the module defining this class if this class is
    # defined by a module *OR* "None" otherwise.
    module_name = get_module_name_qualified_or_none(cls)

    # Return either...
    return (
        # The concatenation of this class and module name if this module name
        # is well-defined.
        '{}.{}'.format(module_name, class_name)
        if module_name is not None else
        # This class name as is otherwise.
        class_name
    )


@type_check
def get_name_unqualified(cls: ClassType) -> str:
    '''
    Unqualified name of the passed class.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this unqualified name for.

    Returns
    ----------
    str
        Unqualified name of this class.
    '''

    # Elegant simplicity diminishes aggressive tendencies.
    return cls.__name__

# ....................{ GETTERS ~ method                  }....................
@type_check
def get_method(cls: ClassType, method_name: str) -> CallableTypes:
    '''
    Possibly unbound method with the passed name defined by the passed class if
    any *or* raise an exception otherwise.

    If this method is:

    * A classmethod, this method is bound to this class.
    * A normal method, this method is unbound.
    * A static method, this method is unbound.

    Parameters
    ----------
    cls : ClassType
        Class to retrieve this method for.
    method_name : str
        Name of the method to be obtained.

    Returns
    ----------
    CallableTypes
        Possibly unbound method with this name.

    Raises
    ----------
    BetseMethodException
        If no such method is defined by this class.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # It's so simple, my eyes bleed.
    return objects.get_callable(obj=cls, callable_name=method_name)

# ....................{ ITERATORS                         }....................
@type_check
def iter_methods(cls: ClassType) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each method defined by
    the passed class (in ascending lexicographic order of method name).

    This generator *only* yields methods statically registered in the internal
    dictionary for this class (e.g., ``__dict__`` in unslotted classes),
    including:

    * Builtin methods, whose names are both prefixed and suffixed by ``__``.
    * Custom methods, whose names are *not* prefixed and suffixed by ``__``,
      including:

      * Custom standard methods.
      * Custom property methods (i.e., methods decorated by the builtin
        :func:`property` decorator).

    Parameters
    ----------
    cls : ClassType
        Class to iterate all methods of.

    Yields
    ----------
    (method_name : str, method : MethodType)
        2-tuple of the name and value of each method bound to this object (in
        ascending lexicographic order of method name).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objiter

    # Well, isn't that special?
    yield from objiter.iter_methods(obj=cls)


@type_check
def iter_methods_matching(
    cls: ClassType, predicate: CallableTypes) -> GeneratorType:
    '''
    Generator yielding 2-tuples of the name and value of each method defined by
    the passed class whose method name matches the passed predicate (in
    ascending lexicographic order of method name).

    Parameters
    ----------
    cls: ClassType
        Class to yield all matching methods of.
    predicate : CallableTypes
        Callable iteratively passed the name of each method and that method
        bound to this object, returning ``True`` only if that name matches this
        predicate.

    Yields
    ----------
    (method_name : str, method : MethodType)
        2-tuple of the name and value of each matching method bound to this
        object (in ascending lexicographic order of method name).

    See Also
    ----------
    :func:`iter_methods`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objiter

    # Well, isn't that special?
    yield from objiter.iter_methods_matching(obj=cls, predicate=predicate)

# ....................{ DEFINERS                          }....................
@type_check
def define_class(
    # Mandatory parameters.
    class_name: str,

    # Optional parameters.
    class_attr_name_to_value: MappingOrNoneTypes = None,
    base_classes: SequenceOrNoneTypes = None,
) -> ClassType:
    '''
    Dynamically define a new class with the passed name subclassing all passed
    base classes and providing all passed class attributes (e.g., class
    variables, methods).

    Parameters
    ----------
    class_name : str
        Name of the class to be created.
    class_attr_name_to_value : MappingType
        Mapping from the name to the initial value of each class attribute
        (e.g., class variable, method) to declare this class to contain.
        Defaults to the empty dictionary, equivalent to declaring a class with
        the trivial body ``pass``.
    base_classes : optional[SequenceTypes]
        Sequence of all base classes to subclass this class from. Defaults to
        the empty tuple, equivalent to the 1-tuple ``(object,)`` containing
        only the root base class of all classes.

    Returns
    ----------
    ClassType
        Class dynamically defined with this name from these base classes and
        class attributes.
    '''

    if class_attr_name_to_value is None:
        class_attr_name_to_value = {}
    if base_classes is None:
        base_classes = ()

    # Thank you, bizarre 3-parameter variant of the type.__init__() method.
    return type(class_name, base_classes, class_attr_name_to_value)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Subcommands accepted by this application's command line interface (CLI).
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and application modules.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from abc import ABCMeta
from betse.util.py import pyident
from betse.util.type.types import (
    type_check, ArgParserType, ArgSubparsersType, SequenceTypes,)

# ....................{ CLASSES ~ container               }....................
class CLISubcommander(object):
    '''
    Container of all CLI subcommands accepted by either a top-level CLI command
    (e.g., ``betse``) or a subcommand of such command (e.g., ``betse plot``).

    This class encapsulates all metadata pertaining to this container,
    including:

    * Human-readable help strings describing this container.
    * All options and arguments accepted by this container.

    Attributes
    ----------
    subcommands : SequenceTypes
        Sequence of all **subcommands** (i.e., :class:`CLISubcommandABC`
        instances) contained by this container.
    subcommand_name_to_arg_parser : MappingType
        Dictionary mapping from the name of each subcommand contained by this
        container to the :class:`ArgumentParser` parsing that subcommand.
    subcommand_name_to_subcommand : MappingType
        Dictionary mapping from the name of each subcommand contained by this
        container to the :class:`CLISubcommandABC` instance describing that
        subcommand.
    subcommand_var_name : str
        Name of the instance variable to which the :class:`ArgumentParser`
        parsing these subcommands from command-line options and arguments
        passed by the external user stores either:

        * The name of the currently passed subcommand for this container if
          the user passed such a subcommand.
        * ``None`` otherwise.
    _help_title : optional[str]
        Human-readable title of the subcommand section in help output,
        typically only one to three words of lowercase unpunctuated text.
    _help_description : str
        Human-readable description to be printed *before* a programmatically
        generated list of these subcommands in help output.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,

        # Mandatory arguments.
        subcommand_var_name: str,
        subcommands: SequenceTypes,

        # Optional arguments.
        help_title: str = 'subcommands',
        help_description: str = '''
Exactly one of the following subcommands must be passed:
''',
    ) -> None:
        '''
        Initialize this container of CLI subcommands.

        Parameters
        ----------
        subcommand_var_name : str
            Name of the instance variable to which the :class:`ArgumentParser`
            parsing these subcommands from command-line options and arguments
            passed by the external user stores either:

            * The name of the currently passed subcommand for this container if
              the user passed such a subcommand.
            * ``None`` otherwise.
        subcommands : SequenceTypes
            Sequence of all **subcommands** (i.e., :class:`CLISubcommandABC`
            instances) to add to this container.
        help_title : optional[str]
            Human-readable title of the subcommand section in help output,
            typically only one to three words of lowercase unpunctuated text.
            For convenience, all format substrings supported by the
            :meth:`CLIABC.expand_help` method (e.g., ``{program_name}``) are
            globally replaced as expected. Defaults to simply ``subcommands``.
        help_description : optional[str]
            Human-readable description to be printed *before* a
            programmatically generated list of these subcommands in help
            output. All format substrings of this string are globally replaced
            as with the ``help_title`` parameter. Defaults to a sensible
            general-purpose string.
        '''

        # Avoid circular import dependencies.
        from betse.util.type.iterable import itertest

        # If this name is *NOT* a valid variable name, raise an exception.
        pyident.die_unless_unqualified(subcommand_var_name)

        # If any passed subcommand is *NOT* a subcommand, raise an exception..
        itertest.die_unless_items_instance_of(
            iterable=subcommands, cls=CLISubcommandABC)

        # Classify all passed parameters.
        self.subcommand_var_name = subcommand_var_name
        self.subcommands = subcommands
        self._help_title = help_title
        self._help_description = help_description

        # Initialize all remaining parameters for safety.
        self.subcommand_name_to_arg_parser = {}
        self.subcommand_name_to_subcommand = {
            subcommand.name: subcommand for subcommand in subcommands}

    # ..................{ ADDERS                            }..................
    @type_check
    def add(
        self,

        # Avoid circular import dependencies.
        cli: 'betse.util.cli.cliabc.CLIABC',
        arg_parser: ArgParserType,
    ) -> ArgSubparsersType:
        '''
        Create a new **argument subparsers container** (i.e., child
        :mod:`argparse`-specific object containing each argument subparser)
        parsing all subcommands in this container, add this container to the
        passed **argument parser** (i.e., parent :mod:`argparse`-specific
        object parsing command-line arguments), and return this container.

        Parameters
        ----------
        cli : betse.util.cli.cliabc.CLIABC
            High-level command-line interface (CLI) owning this container. To
            avoid circular references, neither this method nor any method
            transitively called by this method retains this reference.
        arg_parser : ArgParserType
            Parent argument parser to which the container of subcommand
            argument subparsers created by this method is added. If this parser
            is:

            * The root argument parser for this CLI, all subcommands in this
              container will be created as top-level subcommands.
            * Any other argument parser, all subcommands in this container will
              be created as child subcommands of another parent subcommand.

        Returns
        ----------
        ArgSubparsersType
            Container of subcommand argument subparsers created by this method.
        '''

        # Keyword arguments with which to initialize this container,
        # interpolating all format substrings in all human-readable arguments.
        kwargs = {
            'dest':        self.subcommand_var_name,
            'title':       cli.expand_help(self._help_title),
            'description': cli.expand_help(self._help_description),
        }

        # Container of subcommand argument subparsers to be returned.
        arg_subparsers = arg_parser.add_subparsers(**kwargs)

        # For each contained subcommand:
        #
        # * Add an argument parser parsing this subcommand to this container of
        #   argument subparsers.
        # * Map this subcommand's name to this parser for subsequent lookup.
        for subcommand in self.subcommands:
            self.subcommand_name_to_arg_parser[subcommand.name] = (
                subcommand.add(cli=cli, arg_subparsers=arg_subparsers))

        # Return this container.
        return arg_subparsers

# ....................{ SUPERCLASSES                      }....................
class CLISubcommandABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **CLI subcommand** (i.e., argument passed to
    the external ``betse`` command signifying a high-level action to perform)
    subclasses.

    This class encapsulates all metadata pertaining to this subcommand,
    including:

    * Human-readable help strings describing this subcommand.
    * All options and arguments accepted by this subcommand.

    Attributes
    ----------
    name : str
        Machine-readable alphanumeric name of this CLI subcommand (e.g.,
        ``plot``), typically only a single word. This name should be suitable
        fur use as a valid Python identifier. The :class:`CLIABC` instance
        passed to the :meth:`add` method is expected to define a method named
        ``_do_{name}`` (i.e., this name prefixed by ``_do_``), implementing
        this subcommand.
    _help_synopsis : str
        Human-readable synopsis of this CLI subcommand, typically only one to
        three lines of lowercase, unpunctuated text.
    _help_description : str
        Human-readable description of this CLI subcommand, typically one to
        several paragraphs of grammatical sentences.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        name: str,
        help_synopsis: str,
        help_description: str,
    ) -> None:
        '''
        Initialize this CLI subcommand.

        Parameters
        ----------
        name : str
            Machine-readable alphanumeric name of this CLI subcommand (e.g.,
            ``plot``), typically only a single word. This name should be
            suitable fur use as a valid Python identifier. The :class:`CLIABC`
            instance passed to the :meth:`add` method is expected to define a
            method named ``_do_{name_snakecase}`` implementing this subcommand,
            where ``{name_snakecase}`` is this name converted into snake_case
            (e.g., ``_do_sim_grn`` given a subcommand name of ``sim-grn``).
        help_synopsis : str
            Human-readable synopsis of this CLI subcommand, typically only one
            to three lines of lowercase, unpunctuated text. For convenience,
            all format substrings supported by the :meth:`CLIABC.expand_help`
            method (e.g., ``{program_name}``) are globally replaced as
            expected.
        help_description : str
            Human-readable description of this CLI subcommand, typically one to
            several paragraphs of grammatical sentences. As in the ``synopsis``
            parameter, all format substrings are globally replaced as expected.
        '''

        # Classify all passed parameters.
        self.name = name
        self._help_synopsis = help_synopsis
        self._help_description = help_description

    # ..................{ ADDERS                            }..................
    @type_check
    def add(
        self,

        # Avoid circular import dependencies.
        cli: 'betse.util.cli.cliabc.CLIABC',
        arg_subparsers: ArgSubparsersType,
    ) -> ArgParserType:
        '''
        Create a new **argument subparser** (i.e., :mod:`argparse`-specific
        object parsing command-line arguments) parsing this subcommand, add
        this subparser to the passed collection of **argument subparsers**
        (i.e., another :mod:`argparse`-specific object cotaining multiple
        subparsers), and return this subparser.

        This subparser is configured to:

        * If this subcommand accepts a configuration filename, require such an
          argument be passed.
        * Else, require no arguments be passed.

        Parameters
        ----------
        cli : betse.util.cli.cliabc.CLIABC
            High-level command-line interface (CLI) owning this subcommand. To
            avoid circular references, neither this method nor any method
            transitively called by this method retains this reference.
        arg_subparsers : ArgSubparsersType
            Collection of sibling subcommand argument parsers to which the
            subcommand argument parser created by this method is added. This
            collection is owned either by:

            * A top-level subcommand (e.g., ``plot``), in which case the
              subcommand created by this method is a child of that subcommand.
            * No subcommand, in which case the subcommand created by this
              method is a top-level subcommand.

        Returns
        ----------
        ArgParserType
            Subcommand argument parser created by this method.
        '''

        # Keyword arguments with which to initialize this container,
        # interpolating all format substrings in all human-readable arguments.
        kwargs = {
            'name':        self.name,
            'help':        cli.expand_help(self._help_synopsis),
            'description': cli.expand_help(self._help_description),
        }

        # Merge these subcommand-specific arguments with all default arguments.
        kwargs.update(cli.arg_parser_kwargs)

        # Create and return this parser, added to this container of subparsers.
        return arg_subparsers.add_parser(**kwargs)

# ....................{ SUBCLASSES                        }....................
class CLISubcommandNoArg(CLISubcommandABC):
    '''
    CLI subcommand accepting *no* passed arguments.
    '''

    pass


class CLISubcommandParent(CLISubcommandABC):
    '''
    CLI subcommand which itself is the parent of one or more CLI subcommands,
    accepting *only* the name of a child subcommand as a passed argument.

    Parameters
    ----------
    subcommander : CLISubcommander
        Container of child subcommands accepted by this parent subcommand.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        subcommander: CLISubcommander,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this CLI subcommand.

        Parameters
        ----------
        subcommander : CLISubcommander
            Container of child subcommands accepted by this parent subcommand.

        All remaining parameters are passed as is to the
        :meth:`CLISubcommandABC.__init__` method.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Classify all remaining parameters.
        self.subcommander = subcommander

    # ..................{ ADDERS                            }..................
    @type_check
    def add(
        self,

        # Avoid circular import dependencies.
        cli: 'betse.util.cli.cliabc.CLIABC',
        *args, **kwargs
    ) -> ArgParserType:

        # Argument parser parsing this subcommand.
        arg_parser = super().add(*args, cli=cli, **kwargs)

        # Add this container of argument subparsers to this parser.
        self.subcommander.add(cli=cli, arg_parser=arg_parser)

        # Return this argument parser.
        return arg_parser


class CLISubcommandYAMLOnly(CLISubcommandABC):
    '''
    CLI subcommand accepting *only* a configuration filename as a passed
    argument.
    '''

    # ..................{ ADDERS                            }..................
    def add(self, *args, **kwargs) -> ArgParserType:

        # Subcommand argument subparser added by our superclass.
        arg_subparser = super().add(*args, **kwargs)

        # Configure this subparser to require a configuration file argument.
        arg_subparser.add_argument(
            'conf_filename',
            metavar='CONFIG_FILE',
            help='simulation configuration file',
        )

        # Return this subparser.
        return arg_subparser



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **queue** (i.e., First-In, First-Out (FIFO) data structure, typically
backed by the standard :class:`collections.deque` type) functionality.
'''

# ....................{ IMPORTS                           }....................
import itertools
from betse.util.type.types import type_check, QueueType

# ....................{ POPPERS                           }....................
@type_check
def pop_left(queue: QueueType, count: int) -> None:
    '''
    Pop (i.e., remove) the passed number of items from the head of the passed
    queue in a modestly efficient manner in-place (i.e., modifying the passed
    queue itself).

    Parameters
    ----------
    queue : QueueType
        Queue to pop this number of items from the head of.
    count : int
        Number of items to pop from the head of this queue.

    See Also
    ----------
    https://stackoverflow.com/questions/9507636/how-can-i-pop-lots-of-elements-from-a-deque#comment12041956_9507840
        StackOverflow comment strongly inspiring this implementation.
    '''

    # The following implementation is assumed but has yet to be profiled to be
    # the optimally efficient solution. Alternatives include a map()-based
    # solution, which commonly profiles slower than comparable starmap()-based
    # solutions in the general case:
    #     map(apply, repeat(queue.popleft, count))
    list(itertools.starmap(queue.popleft, itertools.repeat((), count)))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Top-level abstract base class of all command line interface (CLI) subclasses.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and application modules.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import sys
from abc import ABCMeta, abstractmethod
from beartype.typing import (
    Collection,
    Optional,
)
from betse.util.io.log import logs
from betse.util.io.log.conf import logconf
from betse.util.py.pyprofile import profile_callable, ProfileType
from betse.util.type import types
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import (
    type_check,
    ArgParserType,
    MappingType,
    SequenceTypes,
    # SequenceOrNoneTypes,
)

# ....................{ SUPERCLASS                         }....................
class CLIABC(object, metaclass=ABCMeta):
    '''
    Top-level abstract base class of all command line interface (CLI)
    subclasses, suitable for use by both CLI and GUI front-ends for BETSE.

    This superclass provides *no* explicit support for subcommands. Only
    concrete subclasses *not* implementing subcommands should directly subclass
    from this superclass. Concrete subclasses implementing subcommands should
    instead subclass the :class`CLISubcommandableABC` superclass.

    Attributes
    ----------
    _arg_list : list
        List of all passed command-line arguments as unparsed raw strings.
    _arg_parser_top : ArgParserType
        Top-level argument parser parsing all top-level options and subcommands
        passed to this application's external CLI command.
    _args : argparse.Namespace
        :mod:`argparse`-specific container of all passed command-line
        arguments.  See "Attributes (_args)" below for further details.
    _exit_status : IntOrNoneTypes
        Exit status with which to exit this application as a byte in the range
        ``[0, 255]``. Defaults to
        :attr:`betse.util.os.command.cmdexit.SUCCESS`. Subclass
        implementations of the :meth:`_do` method may explicitly override this
        default on failure as a CLI-oriented alternative to exception handling.
        Note that, although exit status is typically returned directly by
        callables, doing so here is infeasible due to the :meth:`_do` method
        API already returning profiled objects.
    _profile_filename : str
        Absolute or relative path of the dumpfile to export a profile of the
        current execution to if :attr:`_profile_type` is
        :attr:`ProfileType.CALL` *or* ignored otherwise.
    _profile_type : ProfileType
        Type of profiling to be performed if any.

    Attributes (of :attr:`_args`)
    ----------
    is_verbose : bool
        ``True`` only if low-level debugging messages are to be logged.
        Defaults to ``False``.
    log_filename : str
        Absolute or relative path of the file to log to. Defaults to the
        absolute path of BETSE's default user-specific logfile.
    log_level : str
        Minimum level of messages to be logged to :attr:`log_filename`,
        formatted as the lowercased name of a :class:`LogLevel` enumeration
        member. Defaults to ``info``.
    stderr_level : str
        Minimum level of messages to be redirected to stderr, formatted as the
        lowercased name of a :class:`LogLevel` enumeration member. Defaults to
        ``warning``.
    stdout_level : str
        Minimum level of messages to be redirected to stdout, formatted as the
        lowercased name of a :class:`LogLevel` enumeration member. Defaults to
        ``info``.
    profile_filename : str
        Absolute or relative path of the dumpfile to export a profile of the
        current execution to if :attr:`profile_type` is ``call`` *or* ignored
        otherwise.  Defaults to the absolute path of BETSE's default
        user-specific profile dumpfile.
    profile_type : str
        Type of profiling to be performed if any, formatted as the lowercased
        name of a :class:`ProfileType` enumeration member. Defaults to
        ``none``.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self):

        # Avoid circular import dependencies.
        from betse.util.os.command.cmdexit import SUCCESS

        # Initialize subclasses performing diamond inheritance if any.
        super().__init__()

        # Default to report success as this application's exit status.
        self._exit_status = SUCCESS

        # For safety, nullify all remaining attributes.
        self._arg_list = None
        self._arg_parser_top = None
        self._arg_parser_kwargs = None
        self._args = None
        self._profile_filename = None
        self._profile_type = None

    # ..................{ RUNNERS                            }..................
    # This method is effectively the main callable of this entire application.
    # This method is thus defined here rather than below, mostly for emphasis.
    @type_check
    def run(self, arg_list: Collection[str] | None = None) -> int:
        '''
        Run the command-line interface (CLI) defined by this subclass with the
        passed argument list if non-``None`` *or* the external argument list
        passed on the command line (i.e., :data:`sys.argv`) otherwise.

        Parameters
        ----------
        arg_list : optional[SequenceTypes]
            Collection of zero or more arguments to pass to this interface.
            Defaults to :data:`None`, in which case arguments passed on the
            command line (i.e., :data:`sys.argv`) are leveraged instead.

        Returns
        -------
        int
            Exit status of this interface in the range ``[0, 255]``.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse.util.os.command.cmdexit import SUCCESS, FAILURE_DEFAULT

        # Default unpassed arguments to those passed on the command line,
        # ignoring the first element of "sys.argv" (i.e., the filename of the
        # command from which the current Python process was spawned).
        if arg_list is None:
            # logs.log_info('Defaulting to sys.argv')
            arg_list = sys.argv[1:]

        #FIXME: Shift the types.is_sequence_nonstr() function into the
        #"betse.util.type.iterable.sequences" submodule.
        assert types.is_sequence_nonstr(arg_list), (
            types.assert_not_sequence_nonstr(arg_list))
        # print('BETSE arg list (in run): {}'.format(arg_list))

        # Classify arguments for subsequent use.
        self._arg_list = arg_list

        try:
            # Parse these arguments *AFTER* initializing logging, ensuring
            # logging of exceptions raised by this parsing.
            self._parse_args()

            # (Re-)initialize all mandatory runtime dependencies of this
            # application *AFTER* parsing and handling all logging-specific CLI
            # options and thus finalizing the logging configuration for the
            # active Python process.
            self._init_app_libs()

            # Run the command-line interface (CLI) defined by this subclass,
            # profiled by the type specified by the "--profile-type" option.
            profile_callable(
                call=self._do,
                profile_type=self._profile_type,
                profile_filename=self._profile_filename,
            )
            # raise ValueError('Test exception handling.')
        except Exception as exception:
            # Handle this exception.
            self._handle_exception(exception)

            # If this application's exit status is still the default and hence
            # has *NOT* been explicitly overriden by the subclass, replace the
            # default status with failure. If this exception provides a
            # system-specific exit status, this status is used; else, the
            # default failure status (i.e., 1) is used.
            #
            # The Windows-specific "winerror" attribute provided by
            # "WindowsError"-based exceptions is ignored. While more
            # fine-grained than the "errno" attribute, "winerror" values are
            # *ONLY* intended to be used internally rather than reported as an
            # exit status to parent processes.
            if self._exit_status == SUCCESS:
                self._exit_status = getattr(
                    exception, 'errno', FAILURE_DEFAULT)

        # Deinitialize this application *AFTER* all prior logic.
        #
        # Note that doing so both nullifies the application metadata singleton
        # and closes open file handles, including those required for logging.
        # For safety, *NO FURTHER APPLICATION LOGIC MAY BE PERFORMED NOW.*
        appmetaone.deinit()

        # Report this application's exit status to the parent process.
        return self._exit_status

    # ..................{ SUBCLASS ~ mandatory               }..................
    # The following methods *MUST* be implemented by subclasses.

    @abstractmethod
    def _do(self) -> object:
        '''
        Implement this command-line interface (CLI) in a subclass-specific
        manner, returning an arbitrary object produced by this logic to be
        memory profiled when the ``--profile-type=size`` CLI option is passed.

        On failure, the subclass implementation of this method should either:

        * Raise an exception, in which case this abstract base class implicitly
          logs this exception and report failure as this application's exit
          status.
        * Explicitly set the :attr:`_exit_status` instance variable to a
          non-zero integer in the range ``[1, 255]`` (e.g.,
          :attr:`betse.util.os.command.cmdexit.FAILURE_DEFAULT`). Note that,
          although exit status is typically returned directly by callables,
          doing so here is infeasible due to this method already returning
          profiled objects.
        '''

        pass

    # ..................{ SUBCLASS ~ mandatory : property    }..................
    # The following properties *MUST* be implemented by subclasses.

    @abstractproperty
    def _help_epilog(self) -> str:
        '''
        Help string template expanded as the **program epilog** (i.e.,
        human-readable string printed after *all* other text in top-level
        application help output).
        '''

        pass

    # ..................{ SUBCLASS ~ optional                }..................
    # The following methods may but need *NOT* be implemented by subclasses.

    def _config_arg_parsing(self) -> None:
        '''
        Configure subclass-specific argument parsing.

        Defaults to a noop.
        '''

        pass

    # ..................{ PROPERTIES : arg parser : kwargs   }..................
    @property
    def arg_parser_kwargs(self) -> MappingType:
        '''
        Dictionary of all keyword arguments to be passed to the
        :meth:`ArgParserType.__init__` and
        :meth:`ArgParserType.add_parser` methods of both the top-level argument
        parser *and* all argument subparsers for this CLI.

        Since the :mod:`argparse` API provides multiple methods rather than a
        single method for creating argument parsers, this versatile dictionary
        is preferred over a monolithic factory-based approach (e.g., a
        hypothetical ``_make_arg_parser_top()`` method).
        '''

        # Avoid circular import dependencies.
        from betse.util.cli.cliarg import SemicolonAwareHelpFormatter

        return {
            # Wrap non-indented lines in help and description text as
            # paragraphs while preserving indented lines in such text as is.
            'formatter_class': SemicolonAwareHelpFormatter,
        }


    @property
    def arg_parser_top_kwargs(self) -> MappingType:
        '''
        Dictionary of all keyword arguments to be passed to the
        :meth:`ArgP"arserType.__init__` method of the top-level argument parser
        for this CLI.

        See Also
        ----------
        :meth:`_init_arg_parser_top`
            Method initializing this argument parser with this dictionary *and*
            keyword arguments common to all argument parsers.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse.util.os.command import cmds

        # Application metadata singleton.
        app_meta = appmetaone.get_app_meta()

        # Dictionary of all keyword arguments to be returned.
        arg_parser_top_kwargs = {
            # Human-readable multi-sentence application description. Since this
            # description is general-purpose rather than CLI-specific, format
            # substrings are *NOT* safely interpolatable into this string.
            'description': app_meta.module_metadata.DESCRIPTION,

            # Human-readable multi-sentence application help suffix.
            'epilog': self.expand_help(self._help_epilog),

            # Basename of the Python wrapper script producing this process.
            'prog': cmds.get_current_basename(),
        }

        # Merge these arguments with all default arguments.
        arg_parser_top_kwargs.update(self.arg_parser_kwargs)

        # Return this dictionary.
        return arg_parser_top_kwargs

    # ..................{ PROPERTIES ~ options               }..................
    @property
    def _options_top(self) -> SequenceTypes:
        '''
        Sequence of all :class:`CLIOptionABC` instances defining the top-level
        CLI options accepted by this application.

        For each such option, a corresponding argument is added to the
        top-level argument parser (i.e., :attr:`_arg_parser_top`).

        **Order is significant,** defining the order that the ``--help`` option
        synopsizes these options in. Options omitted here are *not* parsed by
        argument parsers and are thus effectively ignored.

        Design
        ----------
        Subclasses requiring subclass-specific options are encouraged to:

        * Override this method by:

          * Calling the superclass implementation.
          * Extending the returned tuple with all subclass-specific options.

        * Override the :meth:`_parse_options_top` method by:

          * Calling the superclass implementation.
          * Handling all subclass-specific instance variables parsed into the
            :attr:`self._args` container from these options.

        Returns
        ----------
        SequenceTypes
            Sequence of all such :class:`CLIOptionABC` instances.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse.util.cli.cliopt import (
            CLIOptionArgEnum,
            CLIOptionArgStr,
            CLIOptionBoolTrue,
            CLIOptionVersion,
        )
        from betse.util.io.log.logenum import LogLevel

        # Application metadata singleton.
        app_meta = appmetaone.get_app_meta()

        # Application metadata submodule.
        app_metadata = app_meta.module_metadata

        # Logging configuration singleton.
        log_config = logconf.get_log_conf()

        # Human-readable version specifier suitable for printing to end users.
        version_output = '{} {}'.format(
            app_metadata.NAME, app_metadata.VERSION)

        # If this version is optionally associated with a human-readable
        # codename, suffix this version specifier by that codename.
        if hasattr(app_metadata, 'CODENAME'):
            version_output += ' ({})'.format(app_metadata.CODENAME)

        # List of all default top-level options to be returned.
        options_top = [
            CLIOptionBoolTrue(
                short_name='-v',
                long_name='--verbose',
                synopsis='print and log all messages verbosely',
            ),

            CLIOptionVersion(
                short_name='-V',
                long_name='--version',
                synopsis='print program version and exit',
                version=version_output,
            ),

            CLIOptionArgStr(
                long_name='--log-file',
                synopsis=(
                    'file to log to '
                    '(defaults to "{default}") '
                ),
                var_name='log_filename',
                default_value=log_config.filename,
            ),

            CLIOptionArgEnum(
                long_name='--log-level',
                synopsis=(
                    'minimum level of messages to log to "--log-file" '
                    '(defaults to "{default}") '
                    '[overridden by "--verbose"]'
                ),
                enum_type=LogLevel,
                enum_default=log_config.file_level,
            ),

            CLIOptionArgEnum(
                long_name='--profile-type',
                synopsis=(
                    'type of profiling to perform (defaults to "{default}"):'
                    '\n;* "none", disabling profiling'
                    '\n;* "call", profiling callables (functions, methods)'
                    '\n;* "size", profiling object sizes (requires "pympler")'
                    # ;* "line", profiling code lines (requires "pprofile")
                ),
                enum_type=ProfileType,
                enum_default=ProfileType.NONE,
            ),

            CLIOptionArgStr(
                long_name='--profile-file',
                synopsis=(
                    'file to profile to unless "--profile-type=none" '
                    '(defaults to "{default}")'
                ),
                var_name='profile_filename',
                default_value=(
                    app_meta.profile_default_filename),
            ),
        ]

        # If conditionally exposing the "--matplotlib-backend" option, do so.
        if self._matplotlib_backend_name_forced is None:
            options_top.append(CLIOptionArgStr(
                long_name='--matplotlib-backend',
                synopsis=(
                    'name of matplotlib backend to use '
                    '(see: "betse info")'
                ),
                var_name='matplotlib_backend_name',
                default_value=None,
            ))
        # Else, log this observation for debuggability.
        else:
            logs.log_debug(
                'Mandating matplotlib backend "%s"...',
                self._matplotlib_backend_name_forced)

        # Return this list.
        return options_top


    @property
    def _matplotlib_backend_name_forced(self) -> Optional[str]:
        '''
        Name of the default :mod:`matplotlib` backend to be initialized at
        application startup *or* :data:`None` if this CLI exposes the
        ``--matplotlib-backend`` option enabling end users to specify the name
        of any :mod:`matplotlib` backend.

        Defaults to :data:`None`.
        '''

        return None

    # ..................{ EXPANDERS                          }..................
    @type_check
    def expand_help(self, text: str, **kwargs) -> str:
        '''
        Interpolate the passed keyword arguments into the passed help string
        template, stripping all prefixing and suffixing whitespace from this
        template.

        For convenience, the following default keyword arguments are
        unconditionally interpolated into this template:

        * ``{script_basename}``, expanding to the basename of the Python
          wrapper script running the current application (e.g., ``betse``).
        * ``{program_name}``, expanding to the human-readable name of this
          application (e.g., ``BETSE``).
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse.util.os.command import cmds
        from betse.util.type.text.string import strs

        # Expand it like Expander.
        return strs.remove_whitespace_presuffix(text.format(
            program_name=appmetaone.get_app_meta().module_metadata.NAME,
            script_basename=cmds.get_current_basename(),
            **kwargs
        ))

    # ..................{ ARGS                               }..................
    def _parse_args(self) -> None:
        '''
        Parse all currently passed command-line arguments.

        In order, this method:

        * Creates and configures an argument parser with sensible defaults.
        * Calls the subclass-specific :meth:`_config_arg_parsing` method,
          defaulting to a noop.
        * Parses all arguments with this parser.
        '''

        # Create and configure all argument parsers.
        self._init_arg_parsers()

        # Parse unnamed string arguments into named, typed arguments.
        self._args = self._arg_parser_top.parse_args(self._arg_list)

        # Parse top-level options globally applicable to *ALL* subcommands.
        self._parse_options_top()

        # Log low-level metadata pertaining to the current application *AFTER*
        # parsing top-level options possibly altering this metadata.
        self._log_metadata()


    def _init_arg_parsers(self) -> None:
        '''
        Create and configure all argument parsers, including both the top-level
        argument parser and all subparsers of that parser.
        '''

        # Create and classify the top-level argument parser.
        self._init_arg_parser_top()

        # Configure subclass-specific argument parsing.
        self._config_arg_parsing()


    def _init_arg_parser_top(self) -> None:
        '''
        Create and classify the top-level argument parser.
        '''

        # Top-level argument parser parsing all top-level options and
        # subcommands passed to the CLI command.
        self._arg_parser_top = ArgParserType(**self.arg_parser_top_kwargs)

        #FIXME: Shift into a new top-level add_arg_parser_options() function of
        #the "cliutil" submodule, passing this function the result of
        #self._options_top().

        # For each top-level option, add an argument parsing this option to this
        # argument subparser.
        for option in self._options_top:
            option.add(self._arg_parser_top)

    # ..................{ OPTIONS                            }..................
    def _parse_options_top(self) -> None:
        '''
        Parse **top-level options** (i.e., options globally applicable to all
        subcommands previously declared by the :meth:`_options_top` property).

        Design
        ------
        Subclasses requiring subclass-specific options are encouraged to
        override this method.
        '''

        # Configure logging options *BEFORE* all remaining options, ensuring
        # proper logging of messages emitted by the latter.
        self._parse_options_top_log()
        self._parse_options_top_profile()


    def _parse_options_top_log(self) -> None:
        '''
        Parse top-level logging options globally applicable to all subcommands.
        '''

        # Avoid circular import dependencies.
        from betse.util.io.log.logenum import LogLevel

        # Singleton logging configuration for the current Python process.
        log_config = logconf.get_log_conf()

        # Configure logging according to the passed options. Note that order of
        # assignment is insignificant here.
        # print('is verbose? {}'.format(self._args.is_verbose))
        log_config.is_verbose = self._args.is_verbose
        log_config.filename = self._args.log_filename
        log_config.file_level = LogLevel[self._args.log_level.upper()]

        # Log (and thus display by default) a human-readable synopsis of
        # metadata associated with this application.
        #
        # Note that this logging is intentionally deferred from the earliest
        # time at which logging could technically be performed (namely, the
        # AppMetaABC.init_sans_libs() method). Why? Because logging requires
        # finalizing the logging configuration, which requires parsing *ALL*
        # command-line options. The disadvantage of this otherwise sane
        # approach is, of course, that this logging is deferred.
        self._log_header()

        # Log all string arguments passed to this command.
        logs.log_debug('Passed argument list: {}'.format(self._arg_list))


    def _parse_options_top_profile(self) -> None:
        '''
        Parse top-level profiling options globally applicable to all
        subcommands.
        '''

        # Classify the passed profiling options, converting the profiling type
        # from a raw lowercase string into a full-fledged enumeration member.
        # See _parse_options_top_log() for further detail on this conversion.
        self._profile_filename = self._args.profile_filename
        self._profile_type = ProfileType[self._args.profile_type.upper()]

    # ..................{ DEPENDENCIES                       }..................
    def _init_app_libs(self) -> None:
        '''
        (Re-)initialize all mandatory runtime dependencies of this application
        *after* parsing and handling all logging-specific CLI options and thus
        finalizing the logging configuration for the active Python process.

        This initialization integrates the custom logging and debugging schemes
        implemented by these dependencies with those implemented by this
        application.

        Design
        ----------
        Defaults to (re-)initializing all mandatory runtime dependencies of
        BETSE. Subclasses overriding this method to perform additional
        initialization must manually call the
        :meth:`betse.util.app.meta.appmetaabc.AppMetaABC.init_libs` method to
        initialize these dependencies.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone

        # Application metadata singleton.
        app_meta = appmetaone.get_app_meta()

        # Name of the matplotlib backend to be initialized. Specifically:
        matplotlib_backend_name = (
            # The value of the "--matplotlib-backend" option...
            self._args.matplotlib_backend_name
            # If this CLI exposes this option; else...
            if self._matplotlib_backend_name_forced is None else
            # The name required by the subclass.
            self._matplotlib_backend_name_forced
        )

        # (Re-)initialize all mandatory runtime dependencies.
        app_meta.init_libs(matplotlib_backend_name=matplotlib_backend_name)

    # ..................{ LOGGERS                            }..................
    def _log_header(self) -> None:
        '''
        Display a human-readable synopsis of this application, typically by
        logging the basename and current version of this application and
        various metadata assisting debugging of end user issues.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse import metadata as betse_metadata

        # Metadata submodule specific to the current application.
        app_metadata = appmetaone.get_app_meta().module_metadata

        # Log this in a manner suitable for downstream applications requiring
        # BETSE as an upstream dependency (e.g., BETSEE).
        logs.log_info(
            'Welcome to <<'
            '{script_name} {script_version} | '
            '{betse_name} {betse_version} | '
            '{betse_codename}'
            '>>.'.format(
                script_name=app_metadata.NAME,
                script_version=app_metadata.VERSION,
                betse_name=betse_metadata.NAME,
                betse_version=betse_metadata.VERSION,
                betse_codename=betse_metadata.CODENAME,
            ))


    def _log_metadata(self) -> None:
        '''
        Log low-level metadata pertaining to the current application.

        Specifically, this method logs (in no particular order):

        * The subclass of the currently registered application singleton.
        * The enabling of the default segmentation fault handler.
        * Whether a testing environment is detected.
        * Whether a headless environment is detected.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse.util.os import displays
        from betse.util.test import tsttest
        from betse.util.type.obj import objects

        # Log this metadata.
        logs.log_debug(
            'Application singleton "%s" established.',
            objects.get_class_name_unqualified(appmetaone.get_app_meta()))
        logs.log_debug(
            'Default segmentation fault handler enabled.')
        logs.log_debug(
            'Testing environment detected: %r', tsttest.is_testing())
        logs.log_debug(
            'Headless environment detected: %r', displays.is_headless())

    # ..................{ EXCEPTIONS                         }..................
    @type_check
    def _handle_exception(self, exception: Exception) -> None:
        '''
        Handle the passed uncaught exception, typically by at least logging and
        optionally displaying this exception's detailed traceback to end users.

        Defaults to merely logging this exception. Subclasses may override this
        method to perform additional exception handling, in which case this
        superclass method should still be called to log exceptions.
        '''

        # Log this exception.
        logs.log_exception(exception)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Top-level abstract base class of all **subcommandable command line interface
(CLI)** (i.e., CLI accepting one or more subcommands) subclasses.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and application modules.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from betse.util.cli.cliabc import CLIABC
from betse.util.cli.clicmd import CLISubcommander, CLISubcommandParent
from betse.util.py import pyident
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.obj import objects
from betse.util.type.types import type_check, ArgParserType

# ....................{ SUBCLASS                           }....................
class CLISubcommandableABC(CLIABC):
    '''
    Top-level abstract base class of all **subcommandable command line
    interface (CLI)** (i.e., CLI accepting one or more subcommands) subclasses,
    suitable for use by both CLI and GUI front-ends for BETSE.

    Unlike the parent :class:`CLIABC` superclass, this superclass provides
    explicit support for subcommands. Concrete subclasses implementing
    subcommands should directly subclass this rather than that superclass.
    '''

    # ..................{ SUBCLASS ~ properties              }..................
    # The following properties *MUST* be implemented by subclasses.

    @abstractproperty
    def _subcommander_top(self) -> CLISubcommander:
        '''
        Container of all top-level subcommands accepted by this CLI command.

        **Order is significant,** defining the order that the ``--help`` option
        synopsizes these subcommands in. Subcommands omitted from this
        container will *not* be parsed by argument subparsers and thus ignored.

        For each such subcommand, the :meth:`_config_arg_parsing` method
        creates and adds a corresponding argument subparser to the lower-level
        container of all top-level argument subparsers.
        '''

        pass

    # ..................{ SUPERCLASS ~ properties            }..................
    @property_cached
    def _help_epilog(self) -> str:

        # Sequence of all top-level subcommands.
        subcommands = self._subcommander_top.subcommands

        # If this CLI accepts *NO* subcommands, return the empty string.
        if not subcommands:
            return ''
        # Else, this CLI accepts at least one subcommand.

        # Return a help string embedding the name of the first such subcommand.
        return '''
subcommand help:

For help with a specific subcommand, pass the "-h" or "--help" option to that
subcommand. For example, for help with the "{subcommand_name}" subcommand, run:

;    {{script_basename}} {subcommand_name} --help
'''.format(subcommand_name=subcommands[0].name)

    # ..................{ SUPERCLASS ~ args                  }..................
    def _config_arg_parsing(self) -> None:

        # Container of all top-level argument subparsers for this application.
        self._subcommander_top.add(cli=self, arg_parser=self._arg_parser_top)

    # ..................{ SUPERCLASS ~ cli                   }..................
    def _do(self) -> object:
        '''
        Implement this command-line interface (CLI).

        If the caller passed a top-level subcommand to this CLI command, this
        method runs this subcommand and returns the result of doing so; else,
        this method prints help and returns the current instance of this CLI.
        '''

        self._run_subcommand(
            arg_parser=self._arg_parser_top,
            subcommander=self._subcommander_top,
            subcommand_method_name_prefix='_do_',
        )


    @type_check
    def _run_subcommand(
        self,
        arg_parser: ArgParserType,
        subcommander: CLISubcommander,
        subcommand_method_name_prefix: str,
    ) -> object:
        '''
        Recursively run the child subcommand parsed by the passed argument
        parser, contained by the passed parent subcommands container, and
        selected by the external caller on the command line and return the
        result of doing so if the caller passed a child subcommand for this
        parent subcommand *or* print help for this parent subcommand and return
        the current instance of this CLI.

        To avoid conflict with the ``_do_``-prefixed names of subcommand
        methods, this method is intentionally *not* prefixed by ``_do_``.

        Parameters
        ----------
        arg_parser : ArgParserType
            Argument parsing parsing the parent subcommand of these child
            subcommands.
        subcommander : CLISubcommander
            Container of these child subcommands.
        subcommand_method_name_prefix : str
            Substring prefixing the names of all subclass methods implementing
            these child subcommands.
        '''

        # Name of the child subcommand for this parent subcommand passed by the
        # external user on the command line if a child subcommand was passed
        # *OR* "None" otherwise.
        subcommand_name = getattr(self._args, subcommander.subcommand_var_name)

        # If no child subcommand was passed...
        if not subcommand_name:
            #,Print help for this parent subcommand. Note that this common case
            #constitutes neither a fatal error nor a non-fatal warning.
            print()
            arg_parser.print_help()

            # Return the current instance of this CLI. While trivial, doing so
            # simplifies memory profiling of this CLI.
            return self
        # Else, a child subcommand was passed.

        # Child subcommand and argument parser corresponding to this name.
        subcommand = self._subcommander_top.subcommand_name_to_subcommand[
            subcommand_name]
        subcommand_arg_parser = (
            self._subcommander_top.subcommand_name_to_arg_parser[
                subcommand_name])

        # Snake_case-style name of this subcommand, suitable for use in
        # constructing the syntactically valid Python method name.
        subcommand_name_snakecase = pyident.sanitize_snakecase(subcommand_name)

        # If this subcommand is itself the parent of child subcommands...
        if isinstance(subcommand, CLISubcommandParent):
            # Append the prefix of the names of all subclass methods
            # implementing these child subcommands with this subcommand's name.
            subcommand_method_name_prefix += subcommand_name_snakecase + '_'

            # Call this function recursively and return the result of doing so.
            return self._run_subcommand(
                arg_parser=subcommand_arg_parser,
                subcommander=subcommand.subcommander,
                subcommand_method_name_prefix=subcommand_method_name_prefix,
            )
        # Else, this subcommand is a leaf rather than parent. In this case,
        # terminate recursion by running this subcommand's subclass method.
        else:
            # Name of the method running this subcommand.
            subcommand_method_name = (
                subcommand_method_name_prefix + subcommand_name_snakecase)

            # Method running this subcommand if this method exists *OR* raise
            # a non-human-readable exception otherwise. Usually, that would be
            # bad. In this case, argument parsing coupled with a reliable class
            # implementation ensures this method to exist.
            subcommand_method = objects.get_callable(
                obj=self, callable_name=subcommand_method_name)

            # Run this subcommand and return the result of doing so (if any).
            return subcommand_method()
